# coding=utf-8
"""2D Polyline"""
from __future__ import division
import math

from ._2d import Base2DIn2D
from .pointvector import Point2D
from .line import LineSegment2D
from ..intersection2d import intersect_line2d, intersect_line2d_infinite, \
    closest_point2d_on_line2d
from .._polyline import _group_vertices


class Polyline2D(Base2DIn2D):
    """2D polyline object.

    Args:
        vertices: A list of Point2D objects representing the vertices of the polyline.
        interpolated: Boolean to note whether the polyline should be interpolated
            between the input vertices when it is translated to other interfaces.
            Note that this property has no bearing on the geometric calculations
            performed by this library and is only present in order to assist with
            display/translation.

    Properties:
        * vertices
        * segments
        * min
        * max
        * center
        * p1
        * p2
        * length
        * is_self_intersecting
        * interpolated
    """
    __slots__ = ('_interpolated', '_segments', '_length', '_is_self_intersecting')

    def __init__(self, vertices, interpolated=False):
        """Initialize Polyline2D."""
        Base2DIn2D.__init__(self, vertices)
        self._interpolated = interpolated
        self._segments = None
        self._length = None
        self._is_self_intersecting = None

    @classmethod
    def from_dict(cls, data):
        """Create a Polyline2D from a dictionary.

        Args:
            data: A python dictionary in the following format.

        .. code-block:: python

            {
                "type": "Polyline2D",
                "vertices": [(0, 0), (10, 0), (0, 10)]
            }
        """
        interp = data['interpolated'] if 'interpolated' in data else False
        return cls(tuple(Point2D.from_array(pt) for pt in data['vertices']), interp)

    @classmethod
    def from_array(cls, point_array):
        """Create a Polyline2D from a nested array of vertex coordinates.

        Args:
            point_array: Nested array of point arrays.
        """
        return Polyline2D(Point2D(*point) for point in point_array)

    @classmethod
    def from_polygon(cls, polygon):
        """Create a closed Polyline2D from a Polygon2D.

        Args:
            polygon: A Polygon2D object to be converted to a Polyline2D.
        """
        return Polyline2D(polygon.vertices + (polygon.vertices[0],))

    @property
    def segments(self):
        """Tuple of all line segments in the polyline."""
        if self._segments is None:
            self._segments = \
                tuple(LineSegment2D.from_end_points(vert, self._vertices[i + 1])
                      for i, vert in enumerate(self._vertices[:-1]))
        return self._segments

    @property
    def length(self):
        """The length of the polyline."""
        if self._length is None:
            self._length = sum([seg.length for seg in self.segments])
        return self._length

    @property
    def p1(self):
        """Starting point of the Polyline2D."""
        return self._vertices[0]

    @property
    def p2(self):
        """End point of the Polyline2D."""
        return self._vertices[-1]

    @property
    def is_self_intersecting(self):
        """Boolean noting whether the polyline has self-intersecting segments."""
        if self._is_self_intersecting is None:
            self._is_self_intersecting = False
            _segs = self.segments
            for i, _s in enumerate(_segs[1: len(_segs) - 1]):
                _skip = (i, i + 1, i + 2)
                _other_segs = [x for j, x in enumerate(_segs) if j not in _skip]
                for _oth_s in _other_segs:
                    if _s.intersect_line_ray(_oth_s) is not None:  # intersection!
                        self._is_self_intersecting = True
                        break
                if self._is_self_intersecting is True:
                    break
        return self._is_self_intersecting

    @property
    def interpolated(self):
        """Boolean noting whether the polyline should be interpolated upon translation.

        Note that this property has no bearing on the geometric calculations
        performed by this library and is only present in order to assist with
        display/translation.
        """
        return self._interpolated

    def is_closed(self, tolerance):
        """Test whether this polyline is closed to within the tolerance.

        Args:
            tolerance: The minimum difference between vertices below which vertices
                are considered the same.
        """
        return self._vertices[0].is_equivalent(self._vertices[-1], tolerance)

    def closest_point(self, point):
        """Get the closest Point2D on this object to another Point2D.

        Args:
            point: A Point2D object to which the closest point on this object
                will be computed.

        Returns:
            Point2D for the closest point on this line/ray to the input point.
        """
        close_pts = [closest_point2d_on_line2d(point, seg) for seg in self.segments]
        pt_dts = [point.distance_to_point(close_pt) for close_pt in close_pts]
        s_pts = [p for _, p in sorted(zip(pt_dts, close_pts), key=lambda pair: pair[0])]
        return s_pts[0]

    def distance_to_point(self, point):
        """Get the minimum distance between this object and the input point.

        Args:
            point: A Point2D object to which the minimum distance will be computed.

        Returns:
            The distance to the input point.
        """
        close_pt = self.closest_point(point)
        return point.distance_to_point(close_pt)

    def remove_colinear_vertices(self, tolerance):
        """Get a version of this polyline without colinear or duplicate vertices.

        Args:
            tolerance: The minimum distance that a vertex can be from a line
                before it is considered colinear.

        Returns:
            A Polyline2D with colinear vertices removed. If the removing of colinear
            vertices creates a shape with two or fewer vertices, then a LineSegment2D
            will be returned (which may be zero length if all vertices in the
            Polyline are equal to one another).
        """
        new_vertices = [self.vertices[0]]  # first vertex is always ok
        skip = 0  # track the number of vertices being skipped/removed
        # loop through vertices and remove all cases of colinear verts
        for i, _v in enumerate(self.vertices[1:-1]):
            _v2, _v1 = self[i - skip], self[i + 2]
            _a = _v2.determinant(_v) + _v.determinant(_v1) + _v1.determinant(_v2)
            b_dist = max(_v2.distance_to_point(_v1), _v2.distance_to_point(_v))
            b_dist = tolerance if b_dist < tolerance else b_dist
            tri_tol = (b_dist * tolerance) / 2  # area of triangle with tolerance height
            if abs(_a) >= tri_tol:  # triangle area > tolerance; not colinear
                new_vertices.append(_v)
                skip = 0
            else:
                skip += 1
        new_vertices.append(self[-1])  # last vertex is always ok
        # count the number of vertices to return the correct type of object
        if len(new_vertices) == 2:
            return LineSegment2D.from_end_points(*new_vertices)
        _new_poly = Polyline2D(new_vertices)
        self._transfer_properties(_new_poly)
        return _new_poly

    def smooth(self, strength=1):
        """Get a smoothed version of this Polyline3D by averaging adjacent vertices.

        Args:
            strength: A number between 0 and 1 to represent the strength of smoothing.
                One represents maximum smoothing. Zero represents no smoothing.
        """
        low_wgt = strength * (1 / 3)
        pt_wgts = (low_wgt, 1 - (low_wgt * 2), low_wgt)
        new_pts = [self._vertices[0]]
        for i in range(1, len(self._vertices) - 1):
            rel_pts = self._vertices[i - 1: i + 2]
            sm_pt = Point2D(
                sum(pt.x * w for pt, w in zip(rel_pts, pt_wgts)),
                sum(pt.y * w for pt, w in zip(rel_pts, pt_wgts))
            )
            new_pts.append(sm_pt)
        new_pts.append(self._vertices[-1])
        return Polyline2D(new_pts)

    def reverse(self):
        """Get a copy of this polyline where the vertices are reversed."""
        _new_poly = Polyline2D(tuple(pt for pt in reversed(self.vertices)))
        self._transfer_properties(_new_poly)
        return _new_poly

    def move(self, moving_vec):
        """Get a polyline that has been moved along a vector.

        Args:
            moving_vec: A Vector2D with the direction and distance to move the polyline.
        """
        _new_poly = Polyline2D(tuple(pt.move(moving_vec) for pt in self.vertices))
        self._transfer_properties(_new_poly)
        return _new_poly

    def rotate(self, angle, origin):
        """Get a polyline that is rotated counterclockwise by a certain angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point2D for the origin around which the point will be rotated.
        """
        _new_poly = Polyline2D(tuple(pt.rotate(angle, origin) for pt in self.vertices))
        self._transfer_properties(_new_poly)
        return _new_poly

    def reflect(self, normal, origin):
        """Get a polyline reflected across a plane with the input normal and origin.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the polyline will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point2D representing the origin from which to reflect.
        """
        _new_poly = Polyline2D(tuple(pt.reflect(normal, origin) for pt in self.vertices))
        self._transfer_properties(_new_poly)
        return _new_poly

    def scale(self, factor, origin=None):
        """Scale a polyline by a factor from an origin point.

        Args:
            factor: A number representing how much the polyline should be scaled.
            origin: A Point2D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0).
        """
        if origin is None:
            _new_poly = Polyline2D(tuple(
                Point2D(pt.x * factor, pt.y * factor) for pt in self.vertices))
        else:
            _new_poly = Polyline2D(tuple(
                pt.scale(factor, origin) for pt in self.vertices))
        _new_poly._interpolated = self._interpolated
        return _new_poly

    def offset(self, distance, check_intersection=False):
        """Offset the polyline by a given distance.

        Note that the resulting shape may be self-intersecting if the distance
        is large enough and the is_self_intersecting property may be used to identify
        these shapes.

        Args:
            distance: The distance that the polyline will be offset. Both positive
                and negative values are accepted with positive values being offset
                to the left of the polyline line and negative values being offset
                to the right of the polyline (starting from the first polyline point
                and continuing down the polyline).
            check_intersection: A boolean to note whether the resulting operation
                should be checked for self intersection and, if so, None will be
                returned instead of the self-intersecting polyline.
        """
        # make sure the offset is not zero
        if distance == 0:
            return self

        # loop through the vertices and get the new offset vectors
        middle_verts = list(self._vertices[1:-1])
        if len(middle_verts) != 1:
            middle_verts = [pt for i, pt in enumerate(middle_verts)
                            if pt != middle_verts[i - 1]]
        all_verts = [self._vertices[0]] + middle_verts + [self._vertices[-1]]
        move_vec_st = self.segments[0].v.rotate(math.pi / 2).normalize() * distance
        move_vecs = [move_vec_st]
        for i, pt in enumerate(middle_verts):
            v1 = all_verts[i] - pt
            v2 = all_verts[i + 2] - pt
            ang = v1.angle_counterclockwise(v2) / 2
            if ang == 0:
                ang = math.pi / 2
            m_vec = v1.rotate(ang).normalize()
            m_dist = -distance / math.sin(ang)
            m_vec = m_vec * m_dist
            move_vecs.append(m_vec)
        move_vec_end = self.segments[-1].v.rotate(math.pi / 2).normalize() * distance
        move_vecs.append(move_vec_end)

        # move the vertices by the offset to create the new Polygon2D
        new_pts = tuple(pt.move(m_vec) for pt, m_vec in zip(all_verts, move_vecs))
        new_poly = Polyline2D(new_pts, self.interpolated)

        # check for self intersection between the moving vectors if requested
        if check_intersection:
            poly_segs = new_poly.segments
            _segs = [LineSegment2D(p, v) for p, v in zip(all_verts, move_vecs)]
            _skip = (0, len(_segs) - 1)
            _other_segs = [x for j, x in enumerate(poly_segs) if j not in _skip]
            for _oth_s in _other_segs:
                if _segs[0].intersect_line_ray(_oth_s) is not None:  # intersection!
                    return None
            for i, _s in enumerate(_segs[1: len(_segs)]):
                _skip = (i, i + 1)
                _other_segs = [x for j, x in enumerate(poly_segs) if j not in _skip]
                for _oth_s in _other_segs:
                    if _s.intersect_line_ray(_oth_s) is not None:  # intersection!
                        return None
        return new_poly

    def intersect_line_ray(self, line_ray):
        """Get the intersections between this polyline and a Ray2D or LineSegment2D.

        Args:
            line_ray: A LineSegment2D or Ray2D or to intersect.

        Returns:
            A list with Point2D objects for the intersections.
            List will be empty if no intersection exists.
        """
        intersections = []
        for _s in self.segments:
            inters = intersect_line2d(_s, line_ray)
            if inters is not None:
                intersections.append(inters)
        return intersections

    def intersect_line_infinite(self, ray):
        """Get the intersections between this polyline and a Ray2D extended infinitely.

        Args:
            ray: A Ray2D or to intersect. This will be extended in both
                directions infinitely for the intersection.

        Returns:
            A list with Point2D objects for the intersections.
            List will be empty if no intersection exists.
        """
        intersections = []
        for _s in self.segments:
            inters = intersect_line2d_infinite(_s, ray)
            if inters is not None:
                intersections.append(inters)
        return intersections

    def to_dict(self):
        """Get Polyline2D as a dictionary."""
        base = {'type': 'Polyline2D',
                'vertices': [pt.to_array() for pt in self.vertices]}
        if self.interpolated:
            base['interpolated'] = self.interpolated
        return base

    def to_array(self):
        """Get a list of lists where each sub-list represents a Point2D vertex."""
        return tuple(pt.to_array() for pt in self.vertices)

    def to_polygon(self, tolerance):
        """Get a Polygon2D derived from this object.

        If the polyline is closed to within the tolerance, the segments of this
        polyline and the resulting polygon will match. Otherwise, an extra
        LineSegment2D will be added to connect the start and end of the polyline.

        Args:
            tolerance: The minimum difference between vertices below which vertices
                are considered the same.
        """
        from .polygon import Polygon2D  # must be imported here to avoid circular import
        if self.is_closed(tolerance):
            return Polygon2D(self._vertices[:-1])
        return Polygon2D(self._vertices)

    @staticmethod
    def join_segments(segments, tolerance):
        """Get an array of Polyline2Ds from a list of LineSegment2Ds.

        Args:
            segments: An array of LineSegment2D objects.
            tolerance: The minimum difference in X, Y, and Z values at which Point2Ds
                are considered equivalent. Segments with points that match within the
                tolerance will be joined.

        Returns:
            An array of Polyline2D and LineSegment2D objects assembled from the
            joined segments.
        """
        # group the vertices that make up polylines
        if len(segments) <= 1:
            return segments
        grouped_verts = _group_vertices(segments, tolerance)

        # create the Polyline2D and LineSegment2D objects
        joined_lines = []
        for v_list in grouped_verts:
            if len(v_list) == 2:
                joined_lines.append(LineSegment2D.from_end_points(v_list[0], v_list[1]))
            else:
                joined_lines.append(Polyline2D(v_list))
        return joined_lines

    def _transfer_properties(self, new_polyline):
        """Transfer properties from this polyline to a new polyline."""
        new_polyline._interpolated = self._interpolated
        new_polyline._length = self._length
        new_polyline._is_self_intersecting = self._is_self_intersecting

    def __copy__(self):
        return Polyline2D(self._vertices, self._interpolated)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return tuple(hash(pt) for pt in self._vertices) + (self._interpolated,)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Polyline2D) and self.__key() == other.__key()

    def __repr__(self):
        return 'Polyline2D ({} vertices)'.format(len(self))



# coding=utf-8
"""Base class for all 1D geometries in 3D space (Ray3D and LineSegment3D)."""
from __future__ import division

from .pointvector import Point3D, Vector3D
from ..intersection3d import closest_point3d_on_line3d, \
    closest_point3d_on_line3d_infinite, intersect_line3d_plane


class Base1DIn3D(object):
    """Base class for all 1D geometries in 3D space (Ray3D and LineSegment3D).

    Args:
        p: A Point2D representing the base.
        v: A Vector2D representing the direction.

    Properties:
        * p
        * v
        * min
        * max
        * center
    """
    __slots__ = ('_p', '_v')

    def __init__(self, p, v):
        """Initialize Base1DIn3D."""
        assert isinstance(p, Point3D), "Expected Point3D. Got {}.".format(type(p))
        assert isinstance(v, Vector3D), "Expected Vector3D. Got {}.".format(type(v))
        self._p = p
        self._v = v

    @classmethod
    def from_dict(cls, data):
        """Create a LineSegment3D/Ray3D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "p": (10, 0, 0),
            "v": (10, 10, 0)
            }
        """
        return cls(Point3D.from_array(data['p']),
                   Vector3D.from_array(data['v']))

    @property
    def p(self):
        """Base point."""
        return self._p

    @property
    def v(self):
        """Direction vector."""
        return self._v

    @property
    def min(self):
        """A Point3D for the minimum bounding box vertex around this geometry."""
        p = self._p
        return Point3D(
            min(p.x, p.x + self.v.x), min(p.y, p.y + self.v.y), min(p.z, p.z + self.v.z))

    @property
    def max(self):
        """A Point3D for the maximum bounding box vertex around this geometry."""
        p = self._p
        return Point3D(
            max(p.x, p.x + self.v.x), max(p.y, p.y + self.v.y), max(p.z, p.z + self.v.z))

    @property
    def center(self):
        """A Point3D for the center of the bounding box around this geometry."""
        p = self._p
        return Point3D(p.x + (self.v.x / 2), p.y + (self.v.y / 2), p.z + (self.v.z / 2))

    def is_parallel(self, line_ray, angle_tolerance):
        """Test whether this object is parallel to another LineSegment3D or Ray3D.

        Args:
            line_ray: Another LineSegment3D or Ray3D for which parallelization
                with this objects will be tested.
            angle_tolerance: The max angle in radians that the direction between
                this object and another can vary for them to be considered
                parallel.
        """
        if self.v.angle(line_ray.v) <= angle_tolerance:
            return True
        elif self.v.angle(line_ray.v.reverse()) <= angle_tolerance:
            return True
        return False

    def is_colinear(self, line_ray, tolerance, angle_tolerance=None):
        """Test whether this object is colinear to another LineSegment3D or Ray3D.

        Args:
            line_ray: Another LineSegment3D or Ray3D for which co-linearity
                with this object will be tested.
            tolerance: The maximum distance between the line_ray and the infinite
                extension of this object for them to be considered colinear.
            angle_tolerance: The max angle in radians that the direction between
                this object and another can vary for them to be considered
                parallel. If None, the angle tolerance will not be used to
                evaluate co-linearity and the lines will only be considered
                colinear if the endpoints of one line are within the tolerance
                distance of the other line. (Default: None).
        """
        if angle_tolerance is not None and \
                not self.is_parallel(line_ray, angle_tolerance):
            return False
        _close_pt = closest_point3d_on_line3d_infinite(self.p, line_ray)
        if self.p.distance_to_point(_close_pt) >= tolerance:
            return False
        return True

    def closest_point(self, point):
        """Get the closest Point3D on this object to another Point3D.

        Args:
            point: A Point3D object to which the closest point on this object
                will be computed.

        Returns:
            Point3D for the closest point on this line/ray to the input point.
        """
        return closest_point3d_on_line3d(point, self)

    def distance_to_point(self, point):
        """Get the minimum distance between this object and the input point.

        Args:
            point: A Point3D object to which the minimum distance will be computed.

        Returns:
            The distance to the input point.
        """
        close_pt = self.closest_point(point)
        return point.distance_to_point(close_pt)

    def intersect_plane(self, plane):
        """Get the intersection between this object and a Plane.

        Args:
            plane: A Plane that will be intersected with this object.

        Returns:
            A Point3D object if the intersection was successful.
            None if no intersection exists.
        """
        return intersect_line3d_plane(self, plane)

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def to_dict(self):
        """Get LineSegment3D/Ray3D as a dictionary."""
        return {'p': self.p.to_array(),
                'v': self.v.to_array()}

    def __copy__(self):
        return self.__class__(self.p, self.v)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self.p), hash(self.v))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Base1DIn3D) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Base1Din3D representation."""
        return 'Base 1D Object (3D Space)'



# coding=utf-8
"""Utilities to convert any Ladybug Geometry dictionary to Python objects.

Note that importing this module will import almost all modules within the
Ladybug_geometry library in order to be able to re-serialize almost any
dictionary produced from the library.
"""
from __future__ import division

from ladybug_geometry.geometry2d import Vector2D, Point2D, Ray2D, \
    LineSegment2D, Arc2D, Polyline2D, Polygon2D, Mesh2D
from ladybug_geometry.geometry3d import Vector3D, Point3D, Ray3D, LineSegment3D, \
    Arc3D, Polyline3D, Polyface3D, Mesh3D, Plane, Face3D, Sphere, Cone, Cylinder


def geometry_dict_to_object(ladybug_geom_dict, raise_exception=True):
    """
    Args:
        ladybug_geom_dict (dict): A dictionary of any Ladybug Geometry object.
        raise_exception (bool): Boolean to note whether an exception should be raised
            if the object is not identified as a part of ladybug_geometry.
            Default: True.

    Returns:
        A Python object derived from the input ladybug_geom_dict.
    """

    lbt_types = {
        'Vector2D': Vector2D,
        'Point2D': Point2D,
        'Ray2D': Ray2D,
        'LineSegment2D': LineSegment2D,
        'Arc2D': Arc2D,
        'Polyline2D': Polyline2D,
        'Polygon2D': Polygon2D,
        'Mesh2D': Mesh2D,
        'Vector3D': Vector3D,
        'Point3D': Point3D,
        'Ray3D': Ray3D,
        'LineSegment3D': LineSegment3D,
        'Arc3D': Arc3D,
        'Polyline3D': Polyline3D,
        'Mesh3D': Mesh3D,
        'Plane': Plane,
        'Polyface3D': Polyface3D,
        'Face3D': Face3D,
        'Sphere': Sphere,
        'Cone': Cone,
        'Cylinder': Cylinder
    }

    # Get the ladybug_geometry object 'Type'
    try:
        obj_type = ladybug_geom_dict['type']
    except KeyError:
        raise ValueError('Ladybug dictionary lacks required "type" key.')

    # Build a new Ladybug Python Object based on the "Type"
    try:
        lbt_class = lbt_types[obj_type]
        return lbt_class.from_dict(ladybug_geom_dict)
    except KeyError:
        if raise_exception:
            raise ValueError(
                '{} is not a recognized ladybug geometry type'.format(obj_type))
        else:
            return None



# coding=utf-8
"""A 2D Directed Graph Network data structure used for splitting polygons.

A directed graph (or digraph) is a graph made up of a set of vertices (aka. nodes)
connected by directed edges, often into a network where each node can have
multiple connections.

This class is used in all operations where a polygon or face is split using
line segments or polylines.

The overall strategies used in this module are inspired by operations performed by
NetworkX but were constructed from scratch without working from any particular
module or class in the package More information on NetworkX can be found here:
https://github.com/networkx/networkx
"""
from __future__ import division
import math

from ladybug_geometry.util import coordinates_hash
from ladybug_geometry.intersection2d import intersect_line_segment2d
from ladybug_geometry.geometry2d import LineSegment2D


class Node(object):
    """A Node within in DirectedGraphNetwork, optionally connected to other Nodes.

    Args:
        pt: A Point2D object for the node
        key: String representation of the Point2D object which accounts for tolerance.
        order: Integer for the order of the Node (based on directed graph propagation).
        adj_lst: List of Node objects that are adjacent to this node.
        exterior: Optional boolean to indicate if the Node is on the exterior of
            the graph. If None, this value can be computed later based on the
            position within the overall graph.

    Properties:
        * pt
        * key
        * adj_lst
        * exterior
        * adj_count
    """
    __slots__ = ('key', 'pt', '_order', 'adj_lst', 'exterior')

    def __init__(self, pt, key, order, adj_lst, exterior=None):
        """Initialize Node."""
        self.pt = pt
        self.key = key
        self._order = order
        self.adj_lst = adj_lst
        self.exterior = exterior

    @property
    def adj_count(self):
        """Number of adjacent nodes"""
        return len(self.adj_lst)

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return isinstance(other, Node) and \
            self.key == other.key

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {}'.format(self._order, self.key)


class DirectedGraphNetwork(object):
    """A 2D Directed Graph Network data structure used for splitting polygons.

    A directed graph (or digraph) is a graph made up of a set of vertices (aka. nodes)
    connected by directed edges, often into a network where each node can have
    multiple connections. This class contains for finding the shortest pathways
    through the graph. It also helps differentiate interior from exterior parts
    of the graph. Typically, interior pathways are bi-directional in the graph
    while exterior pathways are uni-directional.

    Args:
        tolerance: The tolerance used to determine point equivalence throughout
            the graph. This is used for hashing points within the network.

    Properties:
        * node_count: Integer for the number of nodes in graph.
        * nodes: An iterable of nodes in graph.
        * ordered_nodes: An iterable of nodes in graph in order they were added.
        * connection_segments: List of LineSegment2D for the node connections.
        * outer_root_node: A node for the outer root key.
        * hole_root_nodes: A list of nodes for the hole root keys.
    """
    __slots__ = ('_directed_graph', '_tolerance', 'outer_root_key', 'hole_root_keys')

    def __init__(self, tolerance):
        """Initialize a PolygonDirectedGraph."""
        self._directed_graph = {}  # will be used to hold the nodes of the network
        # multiply tolerance by 2 to catch both positive and negative point equivalence
        self._tolerance = tolerance * 2
        self.outer_root_key = None  # will be set during network creation
        self.hole_root_keys = []  # will be set during network creation

    @classmethod
    def from_point_array(cls, point_array, tolerance, loop=True):
        """Create a DirectedGraphNetwork for a 1-dimensional array of points.

        Args:
            point_array: Array of Point2D objects.
            tolerance: The tolerance used to determine point equivalence throughout
                the graph. This is used for hashing points within the network.
            loop: Optional boolean, which will ensure that the input point_array
                is connected into a loop. (Default: True).
        """
        ext = True if loop else None
        dg = cls(tolerance)
        for i in range(len(point_array) - 1):
            k = dg.add_node(point_array[i], [point_array[i + 1]], exterior=ext)
            if i == 0:
                dg.outer_root_key = k
        if loop:
            dg.add_node(point_array[-1], [point_array[0]], exterior=ext)
        return dg

    @classmethod
    def from_polygon(cls, polygon, tolerance):
        """Create a DirectedGraphNetwork from a polygon.

        Args:
            polygon: A Polygon2D object.
            tolerance: The tolerance used to determine point equivalence throughout
                the graph. This is used for hashing points within the network.

        Returns:
            A PolygonDirectedGraph object that represents the polygon. The edges
            are uni-directional and counterclockwise. The nodes have the exterior
            property set to True.
        """
        # ensure the boundary and holes are oriented correctly for the graph
        if polygon.is_clockwise:
            polygon = polygon.reverse()
        return cls.from_point_array(polygon.vertices, tolerance, loop=True)

    @classmethod
    def from_shape_with_holes(cls, boundary, holes, tolerance):
        """Create a DirectedGraphNetwork for a shape with holes.

        Args:
            boundary: A Polygon2D for the boundary around the shape.
            holes: An optional list of Polygon2D for the holes within the shape.
                If None, it will be assumed that no holes exist in the shape.
            tolerance: The tolerance used to determine point equivalence throughout
                the graph. This is used for hashing points within the network.

        Returns:
            A PolygonDirectedGraph object that represents the boundary with holes.
            The edges (including the boundary and holes) are uni-directional with
            the outer boundary being counterclockwise and the holes being clockwise.
            In other words, the fill of the shape is always to the left of each edge.
            The nodes have the exterior property set to True.
        """
        # ensure the boundary and holes are oriented correctly for the graph
        if boundary.is_clockwise:
            boundary = boundary.reverse()
        loops = [boundary.vertices]
        if holes is not None:
            for hole in holes:
                if hole.is_clockwise:
                    loops.append(hole.vertices)
                else:
                    loops.append(hole.reverse().vertices)

        # make the directed graph and add the nodes for the boundary + holes
        dg = cls(tolerance=tolerance)
        for loop_count, vertices in enumerate(loops):
            for j in range(len(vertices) - 1):
                curr_v = vertices[j]
                next_v = vertices[j + 1]
                k = dg.add_node(curr_v, [next_v], exterior=True)
                if j == 0:
                    if loop_count == 0:
                        dg.outer_root_key = k
                    else:
                        dg.hole_root_keys.append(k)
            dg.add_node(vertices[-1], [vertices[0]], exterior=True)  # close loop
        return dg

    @classmethod
    def from_shape_to_split(cls, boundary, holes, split_segments, tolerance):
        """Get a DirectedGraphNetwork for a shape to be split with segments.

        The shape is composed of a boundary with optional holes and the
        split_segments are an array of any line segments to be used to
        split the shape.

        Args:
            boundary: A Polygon2D for the boundary around the shape.
            holes: An optional list of Polygon2D for the holes within the shape.
                If None, it will be assumed that no holes exist in the shape.
            split_segments: An array of LineSegment2D to be used to split the shape.
            tolerance: The tolerance used to determine point equivalence throughout
                the graph. This is used for hashing points within the network.

        Returns:
            A PolygonDirectedGraph object that represents the network formed by
            the boundary, holes, and split segments. Portions of the split_segments
            that are outside of the boundary are excluded from the graph. All interior
            connections between the split_segments in the graph are bi-directional.
            The exterior edges (including the boundary and holes) are uni-directional
            with the outer boundary being counterclockwise and the holes being
            clockwise. In other words, the fill of the shape is always to the left
            of each exterior edge. The nodes at the boundary and the holes have
            the exterior property set to True.
        """
        # first split the boundary and holes with the split_segments
        boundary = boundary.remove_colinear_vertices(tolerance)
        bound_sgs = cls._intersect_segments(boundary.segments, split_segments, tolerance)
        bound_pts = [seg.p1 for seg in bound_sgs]
        split_boundary = boundary.__class__(bound_pts)
        split_holes = None
        if holes is not None:
            split_holes = []
            for hole in holes:
                try:
                    hole = hole.remove_colinear_vertices(tolerance)
                    hole_sgs = cls._intersect_segments(
                        hole.segments, split_segments, tolerance)
                    hole_pts = [seg.p1 for seg in hole_sgs]
                    split_holes.append(boundary.__class__(hole_pts))
                except AssertionError:
                    pass  # degenerate hole to ignore
            if len(split_holes) == 0:
                split_holes = None

        # make the directed graph for the boundary + holes
        dg = cls.from_shape_with_holes(split_boundary, split_holes, tolerance)

        # process the split_segments for intersection
        add_segs = list(boundary.segments)
        if holes is not None:
            for hole in holes:
                add_segs.extend(hole.segments)
        split_seg = cls._intersect_segments(split_segments, add_segs, tolerance)
        split_seg = cls._remove_segments_outside_boundary(split_seg, boundary, tolerance)
        if len(split_seg) == 0:  # none of the segments are inside the shape
            return dg

        # add the intersection segments to the graph
        for seg in split_seg:  # add a bidirectional edge to represent interior edges
            dg.add_node(seg.p2, [seg.p1], exterior=False)
            dg.add_node(seg.p1, [seg.p2], exterior=False)
        return dg

    @property
    def node_count(self):
        return len(self.nodes)

    @property
    def nodes(self):
        """Get an iterable of pt nodes"""
        return self._directed_graph.values()

    @property
    def ordered_nodes(self):
        """Get an iterable of pt nodes in order of addition"""
        nodes = list(self.nodes)
        nodes.sort(key=lambda v: v._order)
        return nodes

    @property
    def outer_root_node(self):
        """Get the node of the outer boundary root."""
        return self.node(self.outer_root_key)

    @property
    def hole_root_nodes(self):
        """Get a list of nodes for the roots of the holes."""
        return [self.node(hole_key) for hole_key in self.hole_root_keys]

    @property
    def connection_segments(self):
        """Get a list of LineSegment2D for the node connections in the graph."""
        traversed = set()
        connections = []
        for node in self.nodes:
            for conn_node in node.adj_lst:
                if (conn_node.key, node.key) not in traversed:
                    conn_seg = LineSegment2D.from_end_points(node.pt, conn_node.pt)
                    connections.append(conn_seg)
                    traversed.add((node.key, conn_node.key))
        return connections

    def node(self, key):
        """Retrieves the node based on passed value.

        Args:
            val: The key for a node in the directed graph.

        Returns:
            The node for the passed key.
        """
        try:
            return self._directed_graph[key]
        except KeyError:
            return None  # broken connection

    def add_adj(self, node, adj_val_lst):
        """Adds nodes to node.adj_lst.

        This method will ensure no repetitions will occur in adj_lst.

        Args:
            node: Node to add adjacencies to.
            adj_val_lst: List of Point2D objects to add as adjacent nodes.
        """
        adj_keys = {n.key: None for n in node.adj_lst}
        adj_keys[node.key] = None
        for adj_val in adj_val_lst:
            adj_key = coordinates_hash(adj_val, self._tolerance)
            if adj_key in adj_keys:
                continue

            self._add_node(adj_key, adj_val, exterior=None)
            adj_keys[adj_key] = None
            node.adj_lst.append(self.node(adj_key))

    def remove_adj(self, node, adj_key_lst):
        """Removes nodes in node.adj_lst.

        Args:
            node: Node to remove adjacencies to.
            adj_val_lst: List of adjacency keys to remove as adjacent nodes.
        """
        node.adj_lst = [n for n in node.adj_lst if n.key not in set(adj_key_lst)]

    def add_node(self, val, adj_lst, exterior=None):
        """Add a node into the PolygonDirectedGraph.

        This method consumes a Point2D, computes its key value, and adds it in the
        graph if it doesn't exist. If it does exist it appends adj_lst to existing pt.

        Args:
            val: A Point2D object.
            adj_lst: A list of Point2D objects adjacent to the node.
            exterior: Optional boolean for whether the Node is exterior.

        Returns:
            The hashed key from the existing or new node.
        """
        key = coordinates_hash(val, self._tolerance)  # get key
        self._add_node(key, val, exterior)  # get node if it exists
        node = self._directed_graph[key]
        self.add_adj(node, adj_lst)  # add the adj_lst to dg
        # if the exterior boolean was passed, change the node attribute
        if exterior is not None:
            node.exterior = exterior
        return node.key

    def _add_node(self, key, val, exterior=None):
        """Helper function for add_node.

        If key doesn't currently exist in the graph, it is added with a new key.
        """
        if key not in self._directed_graph:
            self._directed_graph[key] = Node(val, key, self.node_count, [], exterior)
        return self._directed_graph[key]

    def insert_node(self, base_node, new_val, next_node, exterior=None):
        """Insert node in the middle of an edge defined by node and next_node.

        Args:
            base_node: Node object to the left.
            new_val:  A Point2D object for the new node in the middle.
            next_node: Node object to the right.
            exterior: Optional boolean for exterior attribute.

        Returns:
            key of new_val node.
        """
        # add new_val as a node, with next_node as an adjacency
        new_key = self.add_node(new_val, [next_node.pt], exterior=exterior)
        # update parent by adding new adjacency, and removing old adjacency
        self.add_adj(base_node, [self.node(new_key).pt])

        # catch the edge case where the new point is coincident to parent or next_point.
        # this occurs when intersection passes through a corner.
        if (new_key == next_node.key) or (new_key == base_node.key):
            return new_key
        self.remove_adj(base_node, [next_node.key])
        return new_key

    def node_exists(self, key):
        """Check if a node is in the graph. True if node in directed graph else False."""
        return key in self._directed_graph

    def pt_exists(self, pt):
        """True if a point (as Point2D) in directed graph exists as node else False.
        """
        return self.node_exists(coordinates_hash(pt, self._tolerance))

    def polygon_exists(self, polygon):
        """Check if a polygon is in the directed graph.

        Args:
            polygons: A Polygon2D object.

        Return:
            True if exists, else False.
        """
        vertices_loop = list(polygon.vertices)
        vertices_loop = vertices_loop + [vertices_loop[0]]

        for i in range(len(vertices_loop) - 1):
            pt1 = vertices_loop[i]
            pt2 = vertices_loop[i + 1]

            if not self.pt_exists(pt1):
                return False

            node1 = self.node(coordinates_hash(pt1, self._tolerance))
            node2 = self.node(coordinates_hash(pt2, self._tolerance))
            if node2.key in [n.key for n in node1.adj_lst]:
                return False

        return True

    def adj_matrix(self):
        """Gets an adjacency matrix of the directed graph where:

        * 1 = adjacency from row node to col node.
        * 0 = no adjacency.

        Returns:
            N x N square matrix where N is number of nodes.
        """
        nodes = self.ordered_nodes
        # initialize a mtx with no adjacencies
        amtx = [[0 for i in range(self.node_count)]
                for j in range(self.node_count)]

        for i in range(self.node_count):
            adj_indices = [adj._order for adj in nodes[i].adj_lst]
            for adj_idx in adj_indices:
                amtx[i][adj_idx] = 1

        return amtx

    def adj_matrix_labels(self):
        """Returns a dictionary where label key corresponds to index in adj_matrix
        and value is node key"""
        return {i: node.key for i, node in enumerate(self.ordered_nodes)}

    def min_cycle(self, base_node, goal_node, ccw_only=False):
        """Identify the shortest interior cycle between two exterior nodes.

        Args:
            base_node: The first exterior node of the edge.
            goal_node: The end exterior node of the cycle that, together with
                the base_node, constitutes an exterior edge.
            ccw_only: A boolean to note whether the search should be limited
                to the counter-clockwise direction only. (Default: False).

        Returns:
            A list of nodes that form a polygon if the cycle exists, else None.
        """
        # set up a queue for exploring the graph
        explored = []
        queue = [[base_node]]
        orig_dir = base_node.pt - goal_node.pt \
            if base_node.key != goal_node.key else None
        # loop to traverse the graph  with the help of the queue
        while queue:
            path = queue.pop(0)
            node = path[-1]
            # make sure that the current node has not been visited
            if node not in explored:
                prev_dir = node.pt - path[-2].pt if len(path) > 1 else orig_dir
                # iterate over the neighbors to determine relevant nodes
                rel_neighbors, rel_angles = [], []
                last_resort_neighbors, last_resort_angles = [], []
                for neighbor in node.adj_lst:
                    if neighbor == goal_node:  # the shortest path was found!
                        path.append(goal_node)
                        return path
                    edge_dir = neighbor.pt - node.pt
                    cw_angle = prev_dir.angle_clockwise(edge_dir * -1) \
                        if prev_dir is not None else math.pi
                    if 1e-5 < cw_angle < (2 * math.pi) - 1e-5:
                        rel_neighbors.append(neighbor)
                        rel_angles.append(cw_angle)
                    else:  # try to avoid back-tracking along the search
                        last_resort_neighbors.append(neighbor)
                        last_resort_angles.append(cw_angle)
                if len(rel_neighbors) == 0:  # back tracking is the only option
                    rel_neighbors = last_resort_neighbors
                    rel_angles = last_resort_angles
                # sort the neighbors by clockwise angle
                if len(rel_neighbors) > 1:
                    rel_neighbors = [n for _, n in sorted(zip(rel_angles, rel_neighbors),
                                                          key=lambda pair: pair[0])]
                # add the relevant neighbors to the path and the queue
                if ccw_only:
                    new_path = list(path)
                    new_path.append(rel_neighbors[0])
                    queue.append(new_path)
                else:  # add all neighbors to the search
                    for neighbor in rel_neighbors:
                        new_path = list(path)
                        new_path.append(neighbor)
                        queue.append(new_path)
                explored.append(node)
        # if we reached the end of the queue, then no path was found
        return None

    def all_min_cycles(self):
        """Get a list of lists where each sub-list is a minimum cycle of Nodes.

        The combination of all min cycles should account for the full area of
        the input shape if the DirectedGraphNetwork was made using any of the
        class methods that work from polygons. If the DirectedGraphNetwork was made
        using the from_shape_to_split method, the resulting cycles here represent
        the input shape split with the split_segments.
        """
        # first, figure out how many loops each node should be a part of
        node_cycle_counts = {}
        for node in self.nodes:
            node_cycle_counts[node.key] = len(node.adj_lst)

        # loop through the nodes until all min cycles have been identified
        all_cycles = []
        iter_count = 0
        max_iter = len(self.nodes)
        remaining_nodes = self.ordered_nodes
        explored_nodes = set()

        while len(remaining_nodes) > 1 and iter_count < max_iter:
            # try to identify two connected nodes which we can use to build a cycle
            cycle_root = remaining_nodes[0]
            next_node = cycle_root  # if we can't find a connected node, connect to self
            ext_cycle = False
            if cycle_root.exterior:  # exterior cycles tend to have clear connections
                next_node = DirectedGraphNetwork.next_exterior_node(cycle_root)
                if next_node is not None:
                    ext_cycle = True
                else:
                    next_node = cycle_root
            if not ext_cycle:  # see if we can connect it to another incomplete node
                for _next_node in cycle_root.adj_lst:
                    if node_cycle_counts[_next_node.key] != 0:
                        next_node = _next_node
                        ext_cycle = True
                        break

            # find the minimum cycle by searching counter-clockwise
            min_cycle = self.min_cycle(next_node, cycle_root, True)

            # if we found a minimum cycle, evaluate its validity by node connections
            if min_cycle is not None and len(min_cycle) >= 3:
                if not ext_cycle:
                    min_cycle.pop(-1)  # take out the last duplicated node
                is_valid_cycle = True
                for node in min_cycle:
                    if node_cycle_counts[node.key] - 1 < 0:  # we are re-traversing
                        is_valid_cycle = False  # not a valid cycle

                # add the valid cycle to the list to be returned
                if is_valid_cycle:
                    for node in min_cycle:
                        node_cycle_counts[node.key] = node_cycle_counts[node.key] - 1
                        if node_cycle_counts[node.key] == 0:  # all cycles for node found
                            for i, r_node in enumerate(remaining_nodes):
                                if r_node.key == node.key:
                                    remaining_nodes.pop(i)
                                    break
                    all_cycles.append(min_cycle)
                    for node in min_cycle:
                        explored_nodes.add(node.key)

            # reorder the remaining nodes so unexplored nodes get prioritized
            if len(remaining_nodes) != 0:
                for j, node in enumerate(remaining_nodes):
                    if node.key not in explored_nodes:
                        break
                remaining_nodes.insert(0, remaining_nodes.pop(j))
            iter_count += 1

        # if we were not able to address all nodes, see if they are all in the same loop
        if len(remaining_nodes) >= 3:
            current_node = remaining_nodes.pop(0)
            current_node_adj = [node.key for node in node.adj_lst]
            last_cycle = [current_node]
            iter_count, max_iter = 0, len(remaining_nodes)
            while len(remaining_nodes) > 0 and iter_count < max_iter:
                for k, node in enumerate(remaining_nodes):
                    if node.key in current_node_adj:
                        current_node = remaining_nodes.pop(k)
                        current_node_adj = [node.key for node in node.adj_lst]
                        last_cycle.append(current_node)
                        break
                iter_count += 1
            if len(last_cycle) > 2:
                all_cycles.append(last_cycle)

        return all_cycles

    def exterior_cycle(self, cycle_root):
        """Compute exterior boundary from a given node.

        This method assumes that exterior edges are naked (unidirectional) and
        interior edges are bidirectional.

        Args:
            cycle_root: Starting Node in exterior cycle.

        Returns:
            List of nodes on exterior if a cycle exists, else None.
        """
        # Get the first exterior edge
        curr_node = cycle_root
        next_node = DirectedGraphNetwork.next_exterior_node(curr_node)
        if not next_node:
            return None

        # loop through the cycle until we get it all or run out of points
        max_iter = self.node_count + 1  # maximum length a cycle can be
        ext_cycle = [curr_node]
        iter_count = 0
        while next_node.key != cycle_root.key:
            ext_cycle.append(next_node)
            next_node = DirectedGraphNetwork.next_exterior_node(next_node)
            if not next_node:
                return None  # we have hit a dead end in the cycle
            iter_count += 1
            if iter_count > max_iter:
                break  # we have gotten stuck in a loop

        return ext_cycle

    def exterior_cycles(self):
        """Get a list of lists where each sub-list is an exterior cycle of Nodes.

        Exterior cycles refer to the cycles of both the boundary and the holes
        of the DirectedGraphNetwork was created using the from_shape_to_split
        class method.
        """
        exterior_poly_lst = []  # list to store cycles
        explored_nodes = set()  # set to note explored exterior nodes
        max_iter = self.node_count + 1  # maximum length a cycle can be

        # loop through all of the nodes of the graph and find cycles
        for root_node in self.ordered_nodes:
            # make a note that the current node has been explored
            explored_nodes.add(root_node.key)  # mark the node as explored
            # get next exterior adjacent node and check that it's valid
            next_node = self.next_exterior_node(root_node)  # mark the node as explored
            is_valid = (next_node is not None) and (next_node.key not in explored_nodes)
            if not is_valid:
                continue
            # make a note that the next node has been explored
            explored_nodes.add(next_node.key)

            # traverse the loop of points until we get back to start or hit a dead end
            exterior_poly = [root_node]
            prev_node = root_node
            iter_count = 0
            while next_node.key != root_node.key:
                exterior_poly.append(next_node)
                explored_nodes.add(next_node.key)  # mark the node as explored
                follow_node = self.next_exterior_node_no_backtrack(
                    next_node, prev_node, explored_nodes)
                prev_node = next_node  # set as the previous node for the next step
                next_node = follow_node
                if next_node is None:
                    break  # we have hit a dead end in the cycle
                iter_count += 1
                if iter_count > max_iter:
                    print('Extraction of core polygons hit an endless loop.')
                    break  # we have gotten stuck in a loop
            exterior_poly_lst.append(exterior_poly)

        # return all of the exterior loops that were found
        return exterior_poly_lst

    @staticmethod
    def next_exterior_node_no_backtrack(node, previous_node, explored_nodes):
        """Get the next exterior node adjacent to the input node.

        This method is similar to the next_exterior_node method but it includes
        extra checks to handle intersections with 3 or more segments in the
        graph exterior cycles. In these cases a set of previously explored_nodes
        is used to ensure that no back-tracking happens over the search of the
        network, which can lead to infinite looping through the graph. Furthermore,
        the previous_node is used to select the pathway with the smallest angle
        difference with the previous direction. This leads the result towards
        minimal polygons with fewer self-intersecting loops.

        Args:
            node: A Node object for which the next node will be returned.
            previous_node: A Node object for the node that came before
                the current one in the loop. This will be used in the event that
                multiple exterior nodes are found connecting to the input node.
                In this case, the exterior node with the smallest angle difference
                with the previous direction will be returned. This leads the
                result towards minimal polygons and away from self-intersecting
                exterior loops like a bowtie.

        Returns:
            Next node that defines exterior edge, or None if all adjacencies are
            bidirectional.
        """
        # loop through the all adjacent nodes and determine if they are exterior
        next_nodes = []
        for _next_node in node.adj_lst:
            if _next_node.exterior:  # user has labeled it as exterior; we're done!
                return _next_node
            elif _next_node.exterior is None:  # don't know if it's interior or exterior
                # if user-assigned attribute isn't defined, check bi-directionality
                if not DirectedGraphNetwork.is_edge_bidirect(node, _next_node):
                    next_nodes.append(_next_node)

        # evaluate whether there is one obvious choice for the next node
        if len(next_nodes) <= 1:
            return next_nodes[0] if len(next_nodes) == 1 else None
        next_nodes = [nn for nn in next_nodes if nn.key not in explored_nodes]
        if len(next_nodes) <= 1:
            return next_nodes[0] if len(next_nodes) == 1 else None

        # if we have multiple exterior nodes, use the previous node to find the best one
        prev_dir = previous_node.pt - node.pt  # yields a vector
        next_angles = []
        for next_node in next_nodes:
            edge_dir = next_node.pt - node.pt  # yields a vector
            next_angles.append(prev_dir.angle(edge_dir * -1))
        sorted_nodes = [n for _, n in sorted(zip(next_angles, next_nodes),
                                             key=lambda pair: pair[0])]
        return sorted_nodes[0]  # return the node making the smallest angle

    @staticmethod
    def next_exterior_node(node):
        """Get the next exterior node adjacent to consumed node.

        If there are adjacent nodes that are labeled as exterior, with True or
        False defining the Node.exterior property, the first of such nodes in
        the adjacency list will be returned as the next one. Otherwise, the
        bi-directionality will be used to determine whether the next node is
        exterior.

        Args:
            node: A Node object for which the next node will be returned.

        Returns:
            Next node that defines exterior edge, or None if all adjacencies are
            bidirectional.
        """
        # loop through the adjacency and find an exterior node
        for _next_node in node.adj_lst:
            if _next_node.exterior:  # user has labeled it as exterior; we're done!
                return _next_node
            elif _next_node.exterior is None:  # don't know if it's interior or exterior
                # if user-assigned attribute isn't defined, check bi-directionality
                if not DirectedGraphNetwork.is_edge_bidirect(node, _next_node):
                    return _next_node
        return None

    @staticmethod
    def is_edge_bidirect(node1, node2):
        """Are two nodes bidirectional.

        Args:
            node1: Node object
            node2: Node object

        Returns:
            True if node1 and node2 are in each other's adjacency list,
            else False.
        """
        return node1.key in (n.key for n in node2.adj_lst) and \
            node2.key in (n.key for n in node1.adj_lst)

    @staticmethod
    def _intersect_segments(segments, additional_segments, tolerance):
        """Intersect a list of LineSegment2D and split them.

        Args:
            segments: A list of LineSegment2D for the segments to be split/intersected.
            additional_segments: A list of additional LineSegment2Ds, which will be
                used to split the input segments but will not be included in the
                output themselves.
            tolerance: The tolerance at which the intersection will be computed.

        Returns:
            A list of LineSegment2D for the input segments split through
            self-intersection and intersection with the additional_segments.
        """
        # make sure that we are working with lists
        if not isinstance(segments, list):
            segments = list(segments)
        if not isinstance(additional_segments, list):
            additional_segments = list(additional_segments)

        # extend segments a little to ensure intersections happen
        under_tol = tolerance * 0.99
        ext_segments = []
        for seg in segments + additional_segments:
            m_v = seg.v.normalize() * under_tol
            ext_seg = LineSegment2D.from_end_points(seg.p1.move(-m_v), seg.p2.move(m_v))
            ext_segments.append(ext_seg)

        # compute all of the intersection points across the segments
        intersect_pts = [[] for _ in segments]
        for i, seg in enumerate(segments):
            try:
                for other_seg in ext_segments[:i] + ext_segments[i + 1:]:
                    int_pt = intersect_line_segment2d(seg, other_seg)
                    if int_pt is None or int_pt.is_equivalent(seg.p1, tolerance) or \
                            int_pt.is_equivalent(seg.p2, tolerance):
                        continue
                    # we have found an intersection point where segments should be split
                    intersect_pts[i].append(int_pt)
            except IndexError:
                pass  # we have reached the end of the list

        # loop through the segments and split them at the intersection points
        split_segments = []
        for seg, split_pts in zip(segments, intersect_pts):
            if len(split_pts) == 0:
                split_segments.append(seg)
            elif len(split_pts) == 1:  # split the segment in two
                int_pt = split_pts[0]
                split_segments.append(LineSegment2D.from_end_points(seg.p1, int_pt))
                split_segments.append(LineSegment2D.from_end_points(int_pt, seg.p2))
            else:  # sort the points along the segment to split it
                pt_dists = [seg.p1.distance_to_point(ipt) for ipt in split_pts]
                sort_obj = sorted(zip(pt_dists, split_pts), key=lambda pair: pair[0])
                sort_pts = [x for _, x in sort_obj]
                sort_pts.append(seg.p2)
                pr_pt = seg.p1
                for s_pt in sort_pts:
                    if not pr_pt.is_equivalent(s_pt, tolerance):
                        split_segments.append(LineSegment2D.from_end_points(pr_pt, s_pt))
                    pr_pt = s_pt

        return split_segments

    @staticmethod
    def _remove_segments_outside_boundary(segments, boundary, tolerance):
        """Remove LineSegment2D that are outside the boundary of the parent shape.

        This can be used to clean up the result after intersection of segments.

        Args:
            segments: A list of LineSegment2D to be filtered for whether they are
                inside the boundary.
            boundary: A Polygon2D for the boundary of the shape. Segments that lie
                outside of this boundary beyond the tolerance will be removed from
                the result.
            tolerance: The tolerance for distinguishing whether skeleton points lie
                outside the boundary.

        Returns:
            A list of LineSegment2D objects with segments removed that outside
            of the boundary.
        """
        clean_segments = []
        for seg in segments:
            p1, p2 = seg.p1, seg.p2
            if boundary.point_relationship(p2, tolerance) >= 0 and \
                    boundary.point_relationship(p1, tolerance) >= 0:
                clean_segments.append(seg)
        return clean_segments

    def __repr__(self):
        """Represent PolygonDirectedGraph."""
        s = ''
        for n in self.ordered_nodes:
            s += '{}, [{}]\n'.format(
                n.pt.to_array(),
                ', '.join([str(_n.pt.to_array()) for _n in n.adj_lst]))
        return s



# coding=utf-8
"""A class that supports the import and export of OBJ data to/from ladybug_geometry.
"""
import os

try:
    from itertools import izip as zip  # python 2
    writemode = 'wb'
except ImportError:
    writemode = 'w'  # python 3

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry3d.pointvector import Vector3D, Point3D


class OBJ(object):
    """A class that supports the import and export of OBJ data to/from ladybug_geometry.

    Note that ladybug_geometry Mesh3D can be easily created from this OBJ by
    taking the vertices and normals.

    Args:
        vertices: A list or tuple of Point3D objects for vertices.
        faces: A list of tuples with each tuple having either 3 or 4 integers.
            These integers correspond to indices within the list of vertices.
        vertex_texture_map: An optional list or tuple of Point2D that align with the
            vertices input. All coordinate values of the Point2D should be between
            0 and 1 and are intended to map to the XY system of images to be mapped
            onto the OBJ mesh. If None, the OBJ file is written without
            textures. (Default: None).
        vertex_normals: An optional list or tuple of Vector3D that align with the
            vertices input and describe the normal vector to be used at each vertex.
            If None, the OBJ file is written without normals. (Default: None).
        vertex_colors: An optional list of colors that align with the vertices input.
            Note that these are written into the OBJ alongside the vertex
            coordinates separately from the texture map. Not all programs support
            importing OBJs with this color information but Rhino does. (Default: None).
        material_structure: A list of tuples where each tuple contains two elements.
            The first is the identifier of a material that is used in the OBJ and
            the second is the index of the face where the application of the new
            material begins. If None, everything will be assumed to have the
            same diffuse material. (Default: None).

    Properties:
        * vertices
        * faces
        * vertex_texture_map
        * vertex_normals
        * vertex_colors
        * material_structure
    """

    __slots__ = (
        '_vertices', '_faces', '_vertex_texture_map', '_vertex_normals',
        '_vertex_colors', '_material_structure'
    )

    def __init__(
            self, vertices, faces, vertex_texture_map=None, vertex_normals=None,
            vertex_colors=None, material_structure=None
        ):
        self._vertices = self._check_vertices_input(vertices)
        self._faces = self._check_faces_input(faces)
        self.vertex_texture_map = vertex_texture_map
        self.vertex_normals = vertex_normals
        self.vertex_colors = vertex_colors
        self.material_structure = material_structure

    @classmethod
    def from_file(cls, file_path):
        """Create an OBJ object from a .obj file.

        Args:
            file_path: Path to an OBJ file as a text string. Note that, if the file
                includes texture mapping coordinates or vertex normals, the number
                of texture coordinates and normals must align with the number of
                vertices to be importable. Nearly all OBJ files follow this standard.
                If any of the OBJ mesh faces contain more than 4 vertices, only
                the first 4 vertices will be counted.
        """
        vertices, faces, vertex_texture_map, vertex_normals, vertex_colors = \
            [], [], [], [], []
        mat_struct = []
        with open(file_path, 'r') as fp:
            for line in fp:
                if line.startswith('#'):
                    continue
                wds = line.split()
                if len(wds) > 0:
                    first_word = wds[0]
                    if first_word == 'v':  # start of a new vertex
                        vert = Point3D(float(wds[1]), float(wds[2]), float(wds[3]))
                        vertices.append(vert)
                        if len(wds) > 4:
                            vertex_colors.append(tuple(wds[4:]))
                    elif first_word == 'f':  # start of a new face
                        face = []
                        for fv in wds[1:]:
                            face.append(int(fv.split('/')[0]) - 1)
                        if len(face) > 4:  # truncate for compatibility with Mesh3D
                            face = face[:4]
                        faces.append(tuple(face))
                    elif first_word == 'vn':  # start of a new vertex normal
                        norm = Vector3D(float(wds[1]), float(wds[2]), float(wds[3]))
                        vertex_normals.append(norm)
                    elif first_word == 'vt':  # start of a new texture coordinate
                        texture = Point2D(float(wds[1]), float(wds[2]))
                        vertex_texture_map.append(texture)
                    elif first_word == 'usemtl':  # start of a new material application
                        mat_struct.append((wds[1], len(faces)))
        return cls(vertices, faces, vertex_texture_map, vertex_normals,
                   vertex_colors, mat_struct)

    @classmethod
    def from_mesh3d(cls, mesh, include_colors=True, include_normals=False):
        """Create an OBJ object from a ladybug_geometry Mesh3D.
        
        If colors are specified on the Mesh3D, they will be correctly transferred
        to the resulting OBJ object as long as include_colors is True.

        Args:
            mesh: A ladybug_geometry Mesh3D object to be converted to an OBJ object.
            include_colors: Boolean to note whether the Mesh3D colors should be
                transferred to the OBJ object. (Default: True).
            include_normals: Boolean to note whether the vertex normals should be
                included in the resulting OBJ object. (Default: False).
        """
        if include_colors and mesh.is_color_by_face:
            # we need to duplicate vertices to preserve colors
            vertices, faces, colors = [], [], []
            v_ct = 0
            for face_verts, col in zip(mesh.face_vertices, mesh.colors):
                vertices.extend(face_verts)
                if len(face_verts) == 4:
                    faces.append((v_ct, v_ct + 1, v_ct + 2, v_ct + 3))
                    colors.extend([col] * 4)
                    v_ct += 4
                else:
                    faces.append((v_ct, v_ct + 1, v_ct + 2))
                    colors.extend([col] * 3)
                    v_ct += 3
            if include_normals:
                msh_norms = mesh.vertex_normals
                vert_normals = []
                for face in mesh.faces:
                    for fi in face:
                        vert_normals.append(msh_norms[fi])
                return cls(vertices, faces, vertex_normals=msh_norms,
                           vertex_colors=colors)
            return cls(vertices, faces, vertex_colors=colors)
        vertex_colors = mesh.colors if include_colors else None
        if include_normals:
            return cls(mesh.vertices, mesh.faces, vertex_normals=mesh.vertex_normals,
                       vertex_colors=vertex_colors)
        return cls(mesh.vertices, mesh.faces, vertex_colors=vertex_colors)

    @classmethod
    def from_mesh3ds(cls, meshes, material_ids=None, include_normals=False):
        """Create an OBJ object from a list of ladybug_geometry Mesh3D.

        Mesh3D colors are ignored when using this method with the assumption that
        materials are used to specify how the meshes should be rendered.

        Args:
            meshes: A list of ladybug_geometry Mesh3D objects to be converted
                into an OBJ object.
            material_ids: An optional list of strings that aligns with the input
                meshes and denote materials assigned to each mesh. This list of
                material IDs will be automatically converted into an efficient
                material_structure for the OBJ object where materials used for
                multiple meshes only include one reference to the material. If
                None, the OBJ will have no material structure. (Default: None).
            include_normals: Boolean to note whether the vertex normals should be
                included in the resulting OBJ object. (Default: False).
        """
        # sort the meshes by material ID to ensure efficient material structure
        if material_ids is not None:
            assert len(material_ids) == len(meshes), 'Length of OBJ material_ids ({}) ' \
                'does not match the length of meshes ({}).'.format(
                    len(material_ids), len(meshes))
            meshes = [x for _, x in sorted(zip(material_ids, meshes))]
            material_ids = sorted(material_ids)

        # gather all vertices, faces, and (optionally) normals together
        vertices, faces, normals, mat_struct = [], [], [], []
        v_count = 0
        if material_ids is not None:
            last_mat = None
            for mesh, mat_id in zip(meshes, material_ids):
                if mat_id != last_mat:
                    mat_struct.append((mat_id, len(faces)))
                    last_mat = mat_id
                vertices.extend(mesh.vertices)
                if include_normals:
                    normals.extend(mesh.vertex_normals)
                if v_count == 0:
                    faces.extend(mesh.faces)
                else:
                    for f in mesh.faces:
                        faces.append(tuple(fi + v_count for fi in f))
                v_count += len(mesh.vertices)
        else:
            for mesh in meshes:
                vertices.extend(mesh.vertices)
                if include_normals:
                    normals.extend(mesh.vertex_normals)
                if v_count == 0:
                    faces.extend(mesh.faces)
                else:
                    for f in mesh.faces:
                        faces.append(tuple(fi + v_count for fi in f))
                v_count += len(mesh.vertices)

        return cls(
            vertices, faces, vertex_normals=normals, material_structure=mat_struct)

    @property
    def vertices(self):
        """Tuple of Point3D for all vertices in the OBJ."""
        return self._vertices

    @property
    def faces(self):
        """Tuple of tuples for all faces in the OBJ."""
        return self._faces

    @property
    def vertex_texture_map(self):
        """Get or set a tuple of Point2D for texture image coordinates for each vertex.

        Will be None if no texture map is assigned.
        """
        return self._vertex_texture_map

    @vertex_texture_map.setter
    def vertex_texture_map(self, value):
        if value is not None:
            assert isinstance(value, (list, tuple)), 'vertex_texture_map should be ' \
                'a list or tuple. Got {}'.format(type(value))
            if isinstance(value, list):
                value = tuple(value)
            if len(value) == 0:
                value = None
            elif len(value) != len(self.vertices):
                raise ValueError(
                    'Number of items in vertex_texture_map ({}) does not match number'
                    'of OBJ vertices ({}).'.format(len(value), len(self.vertices)))
            else:
                for vert in value:
                    assert isinstance(vert, Point2D), 'Expected Point2D for OBJ ' \
                        'vertex texture. Got {}.'.format(type(vert))
        self._vertex_texture_map = value

    @property
    def vertex_normals(self):
        """Get or set a tuple of Vector3D for vertex normals.

        Will be None if no vertex normals are assigned.
        """
        return self._vertex_normals

    @vertex_normals.setter
    def vertex_normals(self, value):
        if value is not None:
            assert isinstance(value, (list, tuple)), \
                'vertex_normals should be a list or tuple. Got {}'.format(type(value))
            if isinstance(value, list):
                value = tuple(value)
            if len(value) == 0:
                value = None
            elif len(value) != len(self.vertices):
                raise ValueError(
                    'Number of OBJ vertex_normals ({}) does not match the number of'
                    ' OBJ vertices ({}).'.format(len(value), len(self.vertices)))
            else:
                for norm in value:
                    assert isinstance(norm, Vector3D), 'Expected Vector3D for OBJ ' \
                        'vertex normal. Got {}.'.format(type(norm))
        self._vertex_normals = value

    @property
    def vertex_colors(self):
        """Get or set a list of colors for the OBJ. Will be None if no colors assigned.
        """
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, value):
        if value is not None:
            assert isinstance(value, (list, tuple)), \
                'vertex_normals should be a list or tuple. Got {}'.format(type(value))
            if isinstance(value, list):
                value = tuple(value)
            if len(value) == 0:
                value = None
            elif len(value) != len(self.vertices):
                raise ValueError(
                    'Number of OBJ vertex_normals ({}) does not match the number of'
                    ' OBJ vertices ({}).'.format(len(value), len(self.vertices)))
        self._vertex_colors = value

    @property
    def material_structure(self):
        """Get or set a tuple of tuples that specify the material structure of the obj.

        Each sub-tuple contains two elements. The first is the identifier of a
        material that is used in the OBJ and the second is the index of the face
        where the application of the new material begins. If None, everything
        will be assumed to have the same diffuse material.
        """
        return self._material_structure

    @material_structure.setter
    def material_structure(self, value):
        if value is not None:
            assert isinstance(value, (list, tuple)), \
                'vertex_normals should be a list or tuple. Got {}'.format(type(value))
            if len(value) == 0:
                value = None
            else:
                for mt in value:
                    assert isinstance(mt, tuple), 'Expected tuple for OBJ material ' \
                        'structure. Got {}.'.format(type(mt))
                    assert len(mt) == 2, 'OBJ material structure must have 2 items. ' \
                        'Got {}.'.format(len(mt))
                    assert isinstance(mt[0], str), 'Expected String for OBJ material ' \
                        'identifier. Got {}.'.format(type(mt[0]))
                    try:
                        self._faces[mt[1]]
                    except IndexError:
                        raise IndexError(
                            'OBJ material index {} does not correspond to any face. '
                            'There are {} faces in the mesh.'.format(
                                mt[1], len(self._faces)))
                    except TypeError:
                        raise TypeError(
                            'OBJ material must use integers to reference faces. '
                            'Got {}.'.format(type(mt[1])))
                value = sorted(value, key=lambda x: x[1])
                value = tuple(value)
        self._material_structure = value

    def to_file(self, folder, name, triangulate_quads=False, include_mtl=False):
        """Write the OBJ object to an ASCII text file.

        Args:
            folder: A text string for the directory where the OBJ will be written.
            name: A text string for the name of the OBJ file. Note that, if an image
                texture is meant to be assigned to this OBJ, the image should have
                the same name as the one input here except with the .mtl extension
                instead of the .obj extension.
            triangulate_quads: Boolean to note whether quad faces should be
                triangulated upon export to OBJ. This may be needed for certain
                software platforms that require the mesh to be composed entirely
                of triangles (eg. Radiance). (Default: False).
            include_mtl: Boolean to note whether an .mtl file should be automatically
                generated from the material structure written next to the .obj
                file in the output folder. All materials in the mtl file will
                be diffuse white, with the assumption that these will be
                customized later. (Default: False).
        """
        # set up a name and folder
        file_name = name if name.lower().endswith('.obj') else '{}.obj'.format(name)
        obj_file = os.path.join(folder, file_name)
        mtl_file = '{}.mtl'.format(name) if not name.lower().endswith('.obj') else \
            '{}.mtl'.format(name[:-4])

        # write everything into the OBJ file
        with open(obj_file, writemode) as outfile:
            # add a comment at the top to note where the OBJ is written from
            outfile.write('# OBJ file written by ladybug geometry\n\n')

            # add material file name if include_mtl is true
            if self._material_structure is not None or include_mtl:
                if include_mtl:
                    outfile.write('mtllib ' + mtl_file + '\n')
                if self._material_structure is None:
                    outfile.write('usemtl diffuse_0\n')

            # loop through the vertices and add them to the file
            if self.vertex_colors is None:
                for v in self.vertices:
                    outfile.write('v {} {} {}\n'.format(v.x, v.y, v.z))
            else:  # write the vertex colors alongside the vertices
                if len(self.vertex_colors[0]) > 3:
                    for v, c in zip(self.vertices, self.vertex_colors):
                        outfile.write(
                            'v {} {} {} {} {} {}\n'.format(
                                v.x, v.y, v.z, c[0], c[1], c[2])
                        )
                else:  # might be a grayscale weight
                    for v, c in zip(self.vertices, self.vertex_colors):
                        outfile.write(
                            'v {} {} {} {}\n'.format(v.x, v.y, v.z, ' '.join(c))
                        )

            # loop through the texture vertices, if present, and add them to the file
            if self.vertex_texture_map is not None:
                for vt in self.vertex_texture_map:
                    outfile.write('vt {} {}\n'.format(vt.x, vt.y))

            # loop through the normals, if present, and add them to the file
            if self.vertex_normals is not None:
                for vn in self.vertex_normals:
                    outfile.write('vn {} {} {}\n'.format(vn.x, vn.y, vn.z))

            # triangulate the faces if requested
            formatted_faces, formatted_mats = self.faces, self.material_structure
            if triangulate_quads:
                formatted_faces = []
                if formatted_mats is None or len(formatted_mats) == 1:
                    for f in self.faces:
                        if len(f) > 3:
                            formatted_faces.append((f[0], f[1], f[2]))
                            formatted_faces.append((f[2], f[3], f[0]))
                        else:
                            formatted_faces.append(f)
                else:
                    mat_ind = [mat[1] for mat in formatted_mats]
                    for i, f in enumerate(self.faces):
                        if len(f) > 3:
                            formatted_faces.append((f[0], f[1], f[2]))
                            formatted_faces.append((f[2], f[3], f[0]))
                            for j, m in enumerate(formatted_mats):
                                if m[1] > i:
                                    mat_ind[j] = mat_ind[j] + 1
                        else:
                            formatted_faces.append(f)
                    formatted_mats = \
                        [(mn[0], mi) for mn, mi in zip(formatted_mats, mat_ind)]

            # loop through the faces and get all lines of text for them
            face_txt = []
            if self.vertex_texture_map is None and self.vertex_normals is None:
                for f in formatted_faces:
                    face_txt.append('f ' + ' '.join(str(fi + 1) for fi in f) + '\n')
            else:
                if self.vertex_texture_map is not None and \
                        self.vertex_normals is not None:
                    f_map = '{0}/{0}/{0}'
                elif self.vertex_texture_map is None and \
                        self.vertex_normals is not None:
                    f_map = '{0}//{0}'
                else:
                    f_map = '{0}/{0}'
                for f in formatted_faces:
                    face_txt.append(
                        'f ' + ' '.join(f_map.format(fi + 1) for fi in f) + '\n'
                    )

            # write the faces into the file with the material structure
            if formatted_mats is not None:  # insert the materials
                for mat in reversed(formatted_mats):
                    face_txt.insert(mat[1], 'usemtl {}\n'.format(mat[0]))
            for f_lin in face_txt:
                outfile.write(f_lin)

        # write the MTL file if requested
        if include_mtl:
            mat_struct = [('diffuse_0', 0)] if self._material_structure is None else \
                self._material_structure
            mtl_fp = os.path.join(folder, mtl_file)
            with open(mtl_fp, writemode) as mtl_f:
                mtl_f.write('# Ladybug Geometry\n')
                for mat in reversed(mat_struct):
                    mtl_str =  \
                        'newmtl {}\n' \
                        'Ka 0.0000 0.0000 0.0000\n' \
                        'Kd 1.0000 1.0000 1.0000\n' \
                        'Ks 0.0000 0.0000 0.0000\n' \
                        'Tf 0.0000 0.0000 0.0000\n' \
                        'd 1.0000\n' \
                        'Ns 0.0000\n'.format(mat[0])
                    mtl_f.write(mtl_str)

        return obj_file

    def _check_vertices_input(self, vertices):
        """Check the input vertices."""
        if not isinstance(vertices, tuple):
            vertices = tuple(vertices)
        for vert in vertices:
            assert isinstance(vert, Point3D), \
                'Expected Point3D for OBJ vertex. Got {}.'.format(type(vert))
        return vertices

    def _check_faces_input(self, faces):
        """Check input faces for correct formatting."""
        if not isinstance(faces, tuple):
            faces = tuple(faces)
        assert len(faces) > 0, 'OBJ mesh must have at least one face.'
        for f in faces:
            assert isinstance(f, tuple), \
                'Expected tuple for Mesh face. Got {}.'.format(type(f))
            assert len(f) >= 3, \
                'OBJ mesh face must have 3 or more vertices. Got {}.'.format(len(f))
            for ind in f:
                try:
                    self._vertices[ind]
                except IndexError:
                    raise IndexError(
                        'mesh face index {} does not correspond to any vertex. There '
                        'are {} vertices in the mesh.'.format(ind, len(self._vertices)))
                except TypeError:
                    raise TypeError(
                        'Mesh face must use integers to reference vertices. '
                        'Got {}.'.format(type(ind)))
        return faces

    def __len__(self):
        return len(self._vertices)

    def __getitem__(self, key):
        return self._vertices[key]

    def __iter__(self):
        return iter(self._vertices)

    def __repr__(self):
        return 'OBJ ({} vertices) ({} faces)'.format(
            len(self._vertices), len(self._faces))



# coding=utf-8
"""3D Vector and 3D Point"""
from __future__ import division

from ..geometry2d.pointvector import Vector2D

import math
import operator


class Vector3D(object):
    """3D Vector object.

    Args:
        x: Number for the X coordinate.
        y: Number for the Y coordinate.
        z: Number for the Z coordinate.

    Properties:
        * x
        * y
        * z
        * magnitude
        * magnitude_squared
        * is_zero
    """
    __slots__ = ('_x', '_y', '_z')

    def __init__(self, x=0, y=0, z=0):
        """Initialize 3D Vector."""
        self._x = self._cast_to_float(x)
        self._y = self._cast_to_float(y)
        self._z = self._cast_to_float(z)

    @classmethod
    def from_dict(cls, data):
        """Create a Vector3D/Point3D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
                "x": 10,
                "y": 0,
                "z": 0
            }
        """
        return cls(data['x'], data['y'], data['z'])

    @classmethod
    def from_array(cls, array):
        """Initialize a Vector3D/Point3D from an array.

        Args:
            array: A tuple or list with three numbers representing the x, y and z
                values of the point.
        """
        return cls(array[0], array[1], array[2])

    @classmethod
    def from_vector2d(cls, vector2d, z=0):
        """Initialize a new Vector3D from an Vector2D and a z value.

        Args:
            line2d: A Vector2D to be used to generate the Vector3D.
            z: A number for the Z coordinate value of the line.
        """
        return cls(vector2d.x, vector2d.y, z)

    @property
    def x(self):
        """Get the X coordinate."""
        return self._x

    @property
    def y(self):
        """Get the Y coordinate."""
        return self._y

    @property
    def z(self):
        """Get the Z coordinate."""
        return self._z

    @property
    def magnitude(self):
        """Get the magnitude of the vector."""
        return self.__abs__()

    @property
    def magnitude_squared(self):
        """Get the magnitude squared of the vector."""
        return self.x ** 2 + self.y ** 2 + self.z ** 2

    @property
    def min(self):
        """Always equal to (0, 0, 0).

        This property exists to help with bounding box calculations.
        """
        return Point3D(0, 0, 0)

    @property
    def max(self):
        """Always equal to (0, 0, 0).

        This property exists to help with bounding box calculations.
        """
        return Point3D(0, 0, 0)

    def is_zero(self, tolerance):
        """Boolean to note whether the vector is within a given zero tolerance.

        Args:
            tolerance: The tolerance below which the vector is considered to
                be a zero vector.
        """
        return abs(self.x) <= tolerance and abs(self.y) <= tolerance and \
            abs(self.z) <= tolerance

    def is_equivalent(self, other, tolerance):
        """Test whether this object is equivalent to another within a certain tolerance.

        Note that if you want to test whether the coordinate values are perfectly
        equal to one another, the == operator can be used.

        Args:
            other: Another Point3D or Vector3D for which geometric equivalency
                will be tested.
            tolerance: The minimum difference between the coordinate values of two
                objects at which they can be considered geometrically equivalent.

        Returns:
            True if equivalent. False if not equivalent.
        """
        return abs(self.x - other.x) <= tolerance and \
            abs(self.y - other.y) <= tolerance and \
            abs(self.z - other.z) <= tolerance

    def normalize(self):
        """Get a copy of the vector that is a unit vector (magnitude=1)."""
        d = self.magnitude
        try:
            return Vector3D(self.x / d, self.y / d, self.z / d)
        except ZeroDivisionError:
            return self.duplicate()

    def reverse(self):
        """Get a copy of this vector that is reversed."""
        return self.__neg__()

    def dot(self, other):
        """Get the dot product of this vector with another."""
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other):
        """Get the cross product of this vector and another vector."""
        return Vector3D(self.y * other.z - self.z * other.y,
                        -self.x * other.z + self.z * other.x,
                        self.x * other.y - self.y * other.x)

    def angle(self, other):
        """Get the smallest angle between this vector and another."""
        try:
            return math.acos(self.dot(other) / (self.magnitude * other.magnitude))
        except ValueError:  # python floating tolerance can cause math domain error
            if self.dot(other) < 0:
                return math.acos(-1)
            return math.acos(1)

    def rotate(self, axis, angle):
        """Get a vector rotated around an axis through an angle.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle in radians.
        """
        return Vector3D._rotate(self, axis, angle)

    def rotate_xy(self, angle):
        """Get a vector rotated counterclockwise in the XY plane by a certain angle.

        Args:
            angle: An angle in radians.
        """
        vec_2 = Vector2D._rotate(self, angle)
        return Vector3D(vec_2.x, vec_2.y, self.z)

    def reflect(self, normal):
        """Get a vector that is reflected across a plane with the input normal vector.

        Args:
            normal: A Vector3D representing the normal vector for the plane across
                which the vector will be reflected. THIS VECTOR MUST BE NORMALIZED.
        """
        return Vector3D._reflect(self, normal)

    def project(self, normal):
        """Get a vector projected into a plane with a given normal.

        Args:
            normal: A Vector3D representing the normal vector of the plane into which
                the plane will be projected. THIS VECTOR MUST BE NORMALIZED.
        """
        return self - normal * self.dot(normal)

    def duplicate(self):
        """Get a copy of this vector."""
        return self.__copy__()

    def to_dict(self):
        """Get Vector3D as a dictionary."""
        return {'type': 'Vector3D',
                'x': self.x,
                'y': self.y,
                'z': self.z}

    def to_array(self):
        """Get Vector3D/Point3D as a tuple of three numbers"""
        return (self.x, self.y, self.z)

    def _cast_to_float(self, value):
        """Ensure that an input coordinate value is a float."""
        try:
            number = float(value)
        except Exception:
            raise TypeError(
                'Coordinates must be numbers. Got {}: {}.'.format(type(value), value))
        return number

    @staticmethod
    def _reflect(vec, normal):
        """Hidden reflection method used by both Point3D and Vector3D."""
        d = 2 * (vec.x * normal.x + vec.y * normal.y + vec.z * normal.z)
        return Vector3D(vec.x - d * normal.x,
                        vec.y - d * normal.y,
                        vec.z - d * normal.z)

    @staticmethod
    def _rotate(vec, axis, angle):
        """Hidden rotation method used by both Point3D and Vector3D."""
        # Adapted from equations published by Glenn Murray.
        # http://inside.mines.edu/~gmurray/ArbitraryAxisRotation/ArbitraryAxisRotation.html
        x, y, z = vec.x, vec.y, vec.z
        u, v, w = axis.x, axis.y, axis.z

        # Extracted common factors for simplicity and efficiency
        r2 = u ** 2 + v ** 2 + w ** 2
        r = math.sqrt(r2)
        ct = math.cos(angle)
        st = math.sin(angle) / r
        dt = (u * x + v * y + w * z) * (1 - ct) / r2
        return Vector3D((u * dt + x * ct + (-w * y + v * z) * st),
                        (v * dt + y * ct + (w * x - u * z) * st),
                        (w * dt + z * ct + (-v * x + u * y) * st))

    def __copy__(self):
        return self.__class__(self.x, self.y, self.z)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.x, self.y, self.z)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, (Vector3D, Point3D)) and \
            self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __nonzero__(self):
        return self.x != 0 or self.y != 0 or self.z != 0

    def __len__(self):
        return 3

    def __getitem__(self, key):
        return (self.x, self.y, self.z)[key]

    def __iter__(self):
        return iter((self.x, self.y, self.z))

    def __add__(self, other):
        # Vector + Point -> Point
        # Vector + Vector -> Vector
        if isinstance(other, Point3D):
            return Point3D(self.x + other.x, self.y + other.y, self.z + other.z)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x + other.x, self.y + other.y, self.z + other.z)
        else:
            raise TypeError('Cannot add {} and {}'.format(
                self.__class__.__name__, type(other)))

    __radd__ = __add__

    def __sub__(self, other):
        # Vector - Point -> Point
        # Vector - Vector -> Vector
        if isinstance(other, Point3D):
            return Point3D(self.x - other.x, self.y - other.y, self.z - other.z)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x - other.x, self.y - other.y, self.z - other.z)
        else:
            raise TypeError('Cannot subtract {} and {}'.format(
                self.__class__.__name__, type(other)))

    def __rsub__(self, other):
        if isinstance(other, (Vector3D, Point3D)):
            return Vector3D(other.x - self.x, other.y - self.y, other.z - self.z)
        else:
            assert hasattr(other, '__len__') and len(other) == 3, \
                'Cannot subtract types {} and {}'.format(
                    self.__class__.__name__, type(other))
            return Vector3D(other.x - self[0], other.y - self[1], other.z - self[2])

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Vector3D(self.x * other, self.y * other, self.z * other)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x * other.x, self.y * other.y, self.z * other.z)
        elif isinstance(other, Point3D):
            return Point3D(self.x * other.x, self.y * other.y, self.z * other.z)
        else:
            raise TypeError('Cannot multiply {} and {}'.format(
                self.__class__.__name__, type(other)))

    __rmul__ = __mul__

    def __div__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector3D(self.x / other, self.y / other, self.z / other)

    def __rdiv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector3D(other / self.x, other / self.y, other / self.z)

    def __floordiv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector3D(operator.floordiv(self.x, other),
                        operator.floordiv(self.y, other),
                        operator.floordiv(self.z, other))

    def __rfloordiv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector3D(operator.floordiv(other, self.x),
                        operator.floordiv(other, self.y),
                        operator.floordiv(other, self.z))

    def __truediv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector3D(operator.truediv(self.x, other),
                        operator.truediv(self.y, other),
                        operator.truediv(self.z, other))

    def __rtruediv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector3D(operator.truediv(other, self.x),
                        operator.truediv(other, self.y),
                        operator.truediv(other, self.z))

    def __neg__(self):
        return Vector3D(-self.x, -self.y, -self.z)

    __pos__ = __copy__

    def __abs__(self):
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Vector3D representation."""
        return 'Vector3D (%.2f, %.2f, %.2f)' % (self.x, self.y, self.z)


class Point3D(Vector3D):
    """3D Point object.

    Args:
        x: Number for the X coordinate.
        y: Number for the Y coordinate.
        z: Number for the Z coordinate.

    Properties:
        * x
        * y
        * z
    """
    __slots__ = ()

    @classmethod
    def from_point2d(cls, point2d, z=0):
        """Initialize a new Point3D from an Point2D and a z value.

        Args:
            line2d: A Point2D to be used to generate the Point3D.
            z: A number for the Z coordinate value of the line.
        """
        return cls(point2d.x, point2d.y, z)

    @property
    def min(self):
        """Always equal to the point itself.

        This property exists to help with bounding box calculations.
        """
        return self

    @property
    def max(self):
        """Always equal to the point itself.

        This property exists to help with bounding box calculations.
        """
        return self

    def move(self, moving_vec):
        """Get a point that has been moved along a vector.

        Args:
            moving_vec: A Vector3D with the direction and distance to move the point.
        """
        return Point3D(self.x + moving_vec.x,
                       self.y + moving_vec.y,
                       self.z + moving_vec.z)

    def rotate(self, axis, angle, origin):
        """Rotate a point by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the point will be rotated.
        """
        return Vector3D._rotate(self - origin, axis, angle) + origin

    def rotate_xy(self, angle, origin):
        """Get a point rotated counterclockwise in the XY plane by a certain angle.

        Args:
            angle: An angle in radians.
            origin: A Point3D for the origin around which the point will be rotated.
        """
        trans_self = self - origin
        vec_2 = Vector2D._rotate(trans_self, angle)
        return Vector3D(vec_2.x, vec_2.y, trans_self.z) + origin

    def reflect(self, normal, origin):
        """Get a point reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector3D representing the normal vector for the plane across
                which the point will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point3D representing the origin from which to reflect.
        """
        return Vector3D._reflect(self - origin, normal) + origin

    def scale(self, factor, origin=None):
        """Scale a point by a factor from an origin point.

        Args:
            factor: A number representing how much the point should be scaled.
            origin: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        if origin is None:
            return Point3D(self.x * factor, self.y * factor, self.z * factor)
        else:
            return (factor * (self - origin)) + origin

    def project(self, normal, origin):
        """Get a point that is projected into a plane with a given normal and origin.

        Args:
            normal: A Vector3D representing the normal vector of the plane into which
                the plane will be projected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point3D representing the origin the plane into which the
                point will be projected.
        """
        trans_self = self - origin
        return self - normal * trans_self.dot(normal)

    def distance_to_point(self, point):
        """Get the distance from this point to another Point3D.

        Args:
            point: A Point3D to which the distance will be computed.
        """
        vec = (self.x - point.x, self.y - point.y, self.z - point.z)
        return math.sqrt(vec[0] ** 2 + vec[1] ** 2 + vec[2] ** 2)

    def distance_to_plane(self, plane):
        """Get the distance from this point to a Plane.

        Args:
            plane: A Plane object to which the distance will be computed.
        """
        return plane.distance_to_point(self)

    def furthest_distance_to_plane(self, plane):
        """Get the distance from this point to a Plane.

        Equivalent to distance_to_plane. Provided for consistency with other
        geometry classes.

        Args:
            plane: A Plane object to which the distance will be computed.
        """
        return self.distance_to_plane(plane)

    def to_dict(self):
        """Get Point3D as a dictionary."""
        return {'type': 'Point3D',
                'x': self.x,
                'y': self.y,
                'z': self.z}

    def __add__(self, other):
        # Point + Vector -> Point
        # Point + Point -> Vector
        if isinstance(other, Point3D):
            return Vector3D(self.x + other.x, self.y + other.y, self.z + other.z)
        elif isinstance(other, Vector3D):
            return Point3D(self.x + other.x, self.y + other.y, self.z + other.z)
        else:
            raise TypeError('Cannot add Point3D and {}'.format(type(other)))

    def __sub__(self, other):
        # Point - Vector -> Point
        # Point - Point -> Vector
        if isinstance(other, Point3D):
            return Vector3D(self.x - other.x, self.y - other.y, self.z - other.z)
        elif isinstance(other, Vector3D):
            return Point3D(self.x - other.x, self.y - other.y, self.z - other.z)
        else:
            raise TypeError('Cannot subtract Point3D and {}'.format(type(other)))

    def __repr__(self):
        """Point3D representation."""
        return 'Point3D (%.2f, %.2f, %.2f)' % (self.x, self.y, self.z)



# coding=utf-8
"""Base class for all Mesh objects."""
from __future__ import division

import sys
if (sys.version_info > (3, 0)):  # python 3
    xrange = range


class MeshBase(object):
    """Base class for all Mesh objects.

    Args:
        vertices: A list or tuple of Point objects for vertices.
        faces: A list of tuples with each tuple having either 3 or 4 integers.
            These integers correspond to indices within the list of vertices.
        colors: An optional list of colors that correspond to either the faces
            of the mesh or the vertices of the mesh. Default is None.

    Properties:
        * vertices
        * faces
        * colors
        * is_color_by_face
        * face_areas
        * area
        * face_centroids
        * face_area_centroids
        * face_vertices
        * vertex_connected_faces
    """
    __slots__ = ('_vertices', '_faces', '_colors', '_is_color_by_face',
                 '_area', '_face_areas', '_face_centroids', '_face_area_centroids',
                 '_vertex_connected_faces',
                 '_edge_indices', '_edge_types', '_edges',
                 '_naked_edges', '_internal_edges', '_non_manifold_edges')

    def __init__(self, vertices, faces, colors=None):
        """Initialize MeshBase."""
        self._vertices = vertices
        self._faces = self._check_faces_input(faces)
        self._is_color_by_face = False  # default if colors is None
        self.colors = colors
        self._area = None
        self._face_areas = None
        self._face_centroids = None
        self._face_area_centroids = None
        self._vertex_connected_faces = None
        self._edge_indices = None
        self._edge_types = None
        self._edges = None
        self._naked_edges = None
        self._internal_edges = None
        self._non_manifold_edges = None

    @property
    def vertices(self):
        """Tuple of all vertices in this geometry."""
        return self._vertices

    @property
    def faces(self):
        """Tuple of all faces in the mesh."""
        return self._faces

    @property
    def colors(self):
        """Get or set a list of colors for the mesh. Will be None if no colors assigned.
        """
        return self._colors

    @colors.setter
    def colors(self, col):
        if col is not None:
            assert isinstance(col, (list, tuple)), \
                'colors should be a list or tuple. Got {}'.format(type(col))
            if isinstance(col, list):
                col = tuple(col)
            if len(col) == len(self.faces):
                self._is_color_by_face = True
            elif len(col) == len(self.vertices):
                self._is_color_by_face = False
            elif len(col) == 0:
                col = None
            else:
                raise ValueError('Number of colors ({}) does not match the number of'
                                 ' mesh faces ({}) nor the number of vertices ({}).'
                                 .format(len(col), len(self.faces), len(self.vertices)))
        self._colors = col

    @property
    def is_color_by_face(self):
        """Boolean for whether colors are face-by-face (True) or vertex-by-vertex (False)

        When colors are assigned to the mesh, this property is linked to the colors
        and cannot be set. However, when no colors are assigned, it can be set
        because it affects how the mesh is translated to external interfaces.
        In such cases, this property notes whether the mesh will be translated
        in a format that can accept face-by-face colors.
        Default is False when no colors are assigned.
        """
        return self._is_color_by_face

    @property
    def area(self):
        """The area of the entire mesh."""
        if self._area is None:
            self._area = sum(self.face_areas)
        return self._area

    @property
    def face_centroids(self):
        """Tuple of face centroids that aligns the faces property.

        These Point3Ds are are the vertex centroids of the faces, which come from
        considering each quad face as being empty but having equal mass at each vertex.

        Because such vertex centroids are computed as the mean of the coordinate
        values, they are fast to compute and a good property to use when generating
        test points that align large meshes predominantly containing triangles
        and/or convex quad faces. For meshes containing concave quad faces, the
        face_area_centroids may produce a more desirable result.
        """
        if self._face_centroids is None:
            _f_cent = []
            for face in self.faces:
                _f_cent.append(self._face_center(tuple(self._vertices[i] for i in face)))
            self._face_centroids = tuple(_f_cent)
        return self._face_centroids

    @property
    def face_area_centroids(self):
        """Tuple of face area centroids that aligns the faces property.

        These Point3Ds are are the area centroids of the faces, which come from
        considering the surface of each quad face as having constant density.

        The area centroid is more intensive to compute compared to the vertex
        centroid available through the face_centroids property. Moreover, the
        area centroid is equal to the vertex centroid for triangles and fairly
        close to the vertex centroid for convex quad faces with moderate aspect ratios.
        Therefore, this property is most useful when working with meshes containing
        concave quad faces or quads with extreme aspect ratios.
        """
        if self._face_area_centroids is None:
            _f_cent = []
            for face in self.faces:
                if len(face) == 3:
                    _f_cent.append(self._tri_face_centroid(face))
                else:
                    _f_cent.append(self._quad_face_centroid(face))
            self._face_area_centroids = tuple(_f_cent)
        return self._face_area_centroids

    @property
    def face_vertices(self):
        """Tuple of mesh faces with each face defined as a tuple of Point3Ds."""
        return tuple(tuple(self._vertices[i] for i in f) for f in self._faces)

    @property
    def vertex_connected_faces(self):
        """Tuple with a tuple for each vertex that lists the indexes of connected faces.
        """
        if self._vertex_connected_faces is None:
            _vert_faces = [[] for _ in xrange(len(self._vertices))]
            for i, face in enumerate(self._faces):
                for j in face:
                    _vert_faces[j].append(i)
            self._vertex_connected_faces = tuple(tuple(vert) for vert in _vert_faces)
        return self._vertex_connected_faces

    def move(self, moving_vec):
        """Get a mesh that has been moved along a vector.

        Args:
            moving_vec: A Vector with the direction and distance to move the mesh.
        """
        _verts = tuple(pt.move(moving_vec) for pt in self.vertices)
        return self._mesh_transform(_verts)

    def reflect(self, normal, origin):
        """Get a mesh reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector representing the normal vector for the plane across
                which the mesh will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point representing the origin from which to reflect.
        """
        _verts = tuple(pt.reflect(normal, origin) for pt in self.vertices)
        return self._mesh_transform(_verts)

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def _check_faces_input(self, faces):
        """Check input faces for correct formatting."""
        if not isinstance(faces, tuple):
            faces = tuple(faces)
        assert len(faces) > 0, 'Mesh must have at least one face.'
        for f in faces:
            assert isinstance(f, tuple), \
                'Expected tuple for Mesh face. Got {}.'.format(type(f))
            assert len(f) == 3 or len(f) == 4, \
                'Mesh face can only have 3 or 4 vertices. Got {}.'.format(len(f))
            for ind in f:
                try:
                    self._vertices[ind]
                except IndexError:
                    raise IndexError(
                        'mesh face index {} does not correspond to any vertex. There '
                        'are {} vertices in the mesh.'.format(ind, len(self._vertices)))
                except TypeError:
                    raise TypeError(
                        'Mesh face must use integers to reference vertices. '
                        'Got {}.'.format(type(ind)))
        return faces

    def _check_face_pattern(self, pattern):
        """Check input pattern for remove faces for compatibility with this mesh."""
        assert isinstance(pattern, (list, tuple)), 'pattern for remove_faces must' \
            ' be a list or tuple. Got {}.'.format(type(pattern))
        assert len(pattern) == len(self.faces), 'Length of pattern for remove_faces'\
            ' ({}) must match the number of faces in the mesh ({}).'.format(
                len(pattern), len(self.faces))

    def _remove_vertices(self, pattern, face_pattern=None):
        """Get new faces, colors, centroids, and areas when removing vertices."""
        assert isinstance(pattern, (list, tuple)), 'pattern for remove_vertices must' \
            ' be a list or tuple. Got {}.'.format(type(pattern))
        assert len(pattern) == len(self), 'Length of pattern for remove_vertices ({})' \
            ' must match the number of vertices in the mesh ({}).'.format(
                len(pattern), len(self))
        _new_verts = []
        _new_faces = []
        # make a dictionary that maps old vertex indices to new ones
        _vdict = {}
        _vcount = 0
        for i, _in_mesh in enumerate(pattern):
            if _in_mesh is True:
                _vdict[i] = _vcount
                _vcount += 1
                _new_verts.append(self._vertices[i])

        # get the new faces
        if face_pattern is None:
            face_pattern = []
            for _f in self.faces:
                try:
                    if len(_f) == 3:
                        _new_f = (_vdict[_f[0]], _vdict[_f[1]], _vdict[_f[2]])
                    else:
                        _new_f = (_vdict[_f[0]], _vdict[_f[1]], _vdict[_f[2]],
                                  _vdict[_f[3]])
                    _new_faces.append(_new_f)
                    face_pattern.append(True)
                except KeyError:
                    face_pattern.append(False)
        else:
            for i, _f in enumerate(self.faces):
                if face_pattern[i] is True:
                    if len(_f) == 3:
                        _new_f = (_vdict[_f[0]], _vdict[_f[1]], _vdict[_f[2]])
                    else:
                        _new_f = (_vdict[_f[0]], _vdict[_f[1]], _vdict[_f[2]],
                                  _vdict[_f[3]])
                    _new_faces.append(_new_f)

        # remove colors if they are assigned
        _new_colors = None
        if self._colors is not None:
            _new_col = []
            if self._is_color_by_face is True:
                for i, _p in enumerate(face_pattern):
                    if _p is True:
                        _new_col.append(self._colors[i])
            else:
                for i, _p in enumerate(pattern):
                    if _p is True:
                        _new_col.append(self._colors[i])
            _new_colors = tuple(_new_col)

        # transfer face centroids and areas if they exist so they don't get re-computed
        _new_f_cent, _new_f_area = self._transfer_face_centroids_areas(face_pattern)

        return tuple(_new_verts), tuple(_new_faces), _new_colors, \
            _new_f_cent, _new_f_area, face_pattern

    def _transfer_face_centroids_areas(self, face_pattern):
        """Get face centroids and areas when removing faces.

        This is so they can be transferred to the new mesh.
        """
        _new_f_cent = None
        if self._face_centroids is not None:
            _new_f_cent = tuple(
                self._face_centroids[i] for i, _p in enumerate(face_pattern) if _p)
        _new_f_area = None
        if self._face_areas is not None:
            if isinstance(self._face_areas, (float, int)):
                _new_f_area = self._face_areas
            else:
                _new_f_area = tuple(
                    self._face_areas[i] for i, _p in enumerate(face_pattern) if _p)
        return _new_f_cent, _new_f_area

    def _vertex_pattern_from_remove_faces(self, pattern):
        """Get a pattern of vertices to remove from a pattern of faces to remove."""
        self._check_face_pattern(pattern)
        vertex_pattern = [False for _v in self.vertices]
        for i, _in_mesh in enumerate(pattern):
            if _in_mesh is True:
                _face = self._faces[i]
                for j in _face:
                    vertex_pattern[j] = True
        return vertex_pattern

    def _remove_faces_only(self, pattern):
        """Get new faces, colors, centroids, and areas when removing faces only."""
        self._check_face_pattern(pattern)
        _new_faces = []
        for i, _in_mesh in enumerate(pattern):
            if _in_mesh is True:
                _new_faces.append(self._faces[i])

        # transfer the colors over to the new mesh if they are assigned face-by-face
        _new_colors = self._colors
        if self._colors is not None and self._is_color_by_face is True:
            _new_col = []
            for i, _p in enumerate(pattern):
                if _p is True:
                    _new_col.append(self._colors[i])
            _new_colors = _new_col

        # transfer face centroids and areas if they exist so they don't get re-computed
        _new_f_cent, _new_f_area = self._transfer_face_centroids_areas(pattern)

        return tuple(_new_faces), _new_colors, _new_f_cent, _new_f_area

    def _compute_edge_info(self):
        """Compute information about the edges of the mesh faces."""
        edge_i = []
        edge_t = []
        for fi in self.faces:
            for i, vi in enumerate(fi):
                try:  # this can get slow for large number of vertices
                    ind = edge_i.index((vi, fi[i - 1]))
                    edge_t[ind] += 1
                except ValueError:  # make sure reversed edge isn't there
                    try:
                        ind = edge_i.index((fi[i - 1], vi))
                        edge_t[ind] += 1
                    except ValueError:  # add a new edge
                        if fi[i - 1] != vi:  # avoid cases of same start and end
                            edge_i.append((fi[i - 1], vi))
                            edge_t.append(0)
        self._edge_indices = edge_i if isinstance(edge_i, tuple) else tuple(edge_i)
        self._edge_types = edge_t if isinstance(edge_t, tuple) else tuple(edge_t)

    def _transfer_properties(self, new_mesh):
        """Transfer properties when making a copy of the mesh or doing transforms."""
        new_mesh._colors = self._colors
        new_mesh._is_color_by_face = self._is_color_by_face
        new_mesh._face_areas = self._face_areas
        new_mesh._area = self._area

    def _transfer_properties_scale(self, new_mesh, factor):
        """Transfer properties when performing a scale operation."""
        new_mesh._colors = self._colors
        new_mesh._is_color_by_face = self._is_color_by_face
        if self._face_areas is not None:
            new_mesh._face_areas = tuple(a * factor for a in self._face_areas)
        if self._area is not None:
            new_mesh._area = self._area * factor

    @staticmethod
    def _interpret_input_from_face_vertices(faces, purge):
        """Get faces and vertices from a list of faces as points."""
        vertices = []  # collection of vertices as point objects
        face_collector = []  # collection of face indices
        if purge:
            for f in faces:
                ind = []
                for v in f:
                    try:  # this can get very slow for large number of vertices.
                        ind.append(vertices.index(v))
                    except ValueError:  # add new point
                        vertices.append(v)
                        ind.append(len(vertices) - 1)
                face_collector.append(tuple(ind))
        else:
            ver_counter = 0
            for f in faces:
                ind = []
                for v in f:
                    vertices.append(v)
                    ind.append(ver_counter)
                    ver_counter += 1
                face_collector.append(tuple(ind))
        return vertices, face_collector

    @staticmethod
    def _interpret_input_from_face_vertices_with_tolerance(faces, tolerance):
        """Get faces and vertices from a list of faces as points.

        This differs from _interpret_input_from_face_vertices in that a tolerance
        will be used to check for vertex equivalency instead of relying on
        equality within floating point tolerance.

        Args:
            faces: A list of faces where each face is a list of points.
            tolerance: A number for the tolerance to use when checking for
                duplicate vertices.

        Returns:
            A tuple of vertices and faces.
        """

        def index_of_equivalent_point(vertix_list, vertix):
            """Get the index within a list of vertices using the is_equivalent test."""
            for i, other_vert in enumerate(vertix_list):
                if vertix.is_equivalent(other_vert, tolerance):
                    return i
            raise ValueError()

        vertices = []
        face_collector = []
        for f in faces:
            ind = []
            for v in f:
                try:
                    ind.append(index_of_equivalent_point(vertices, v))
                except ValueError:  # add new point
                    vertices.append(v)
                    ind.append(len(vertices) - 1)
            face_collector.append(tuple(ind))
        return vertices, face_collector

    def __len__(self):
        return len(self._vertices)

    def __getitem__(self, key):
        return self._vertices[key]

    def __iter__(self):
        return iter(self._vertices)

    def __copy__(self):
        return MeshBase(self._vertices, self._faces, self._colors)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return tuple(hash(pt) for pt in self._vertices) + \
            tuple(hash(face) for face in self._faces)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, MeshBase) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Base MEsh representation."""
        return 'Base Mesh Object'



# coding=utf-8
"""Base class for all 1D geometries in 2D space (Ray2D and LineSegment2D)."""
from __future__ import division

from .pointvector import Vector2D, Point2D
from ..intersection2d import intersect_line2d, closest_point2d_on_line2d, \
    closest_point2d_on_line2d_infinite


class Base1DIn2D(object):
    """Base class for all 1D geometries in 2D space (Ray2D and LineSegment2D).

    Args:
        p: A Point2D representing the base.
        v: A Vector2D representing the direction.

    Properties:
        * p
        * v
        * min
        * max
        * center
    """
    __slots__ = ('_p', '_v')

    def __init__(self, p, v):
        """Initialize Base1DIn2D."""
        assert isinstance(p, Point2D), "Expected Point2D. Got {}.".format(type(p))
        assert isinstance(v, Vector2D), "Expected Vector2D. Got {}.".format(type(v))
        self._p = p
        self._v = v

    @classmethod
    def from_dict(cls, data):
        """Create a LineSegment2D/Ray2D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "p": (10, 0),
            "v": (10, 10)
            }
        """
        return cls(Point2D.from_array(data['p']),
                   Vector2D.from_array(data['v']))

    @property
    def p(self):
        """Base point."""
        return self._p

    @property
    def v(self):
        """Direction vector."""
        return self._v

    @property
    def min(self):
        """A Point2D for the minimum bounding rectangle vertex around this geometry."""
        p = self._p
        return Point2D(min(p.x, p.x + self.v.x), min(p.y, p.y + self.v.y))

    @property
    def max(self):
        """A Point2D for the maximum bounding rectangle vertex around this geometry."""
        p = self._p
        return Point2D(max(p.x, p.x + self.v.x), max(p.y, p.y + self.v.y))

    @property
    def center(self):
        """A Point2D for the center of the bounding rectangle around this geometry."""
        p = self._p
        return Point2D(p.x + (self.v.x / 2), p.y + (self.v.y / 2))

    def closest_point(self, point):
        """Get the closest Point2D on this object to another Point2D.

        Args:
            point: A Point2D object to which the closest point on this object
                will be computed.

        Returns:
            Point2D for the closest point on this line to the input point.
        """
        return closest_point2d_on_line2d(point, self)

    def distance_to_point(self, point):
        """Get the minimum distance between this object and the input point.

        Args:
            point: A Point2D object to which the minimum distance will be computed.

        Returns:
            The distance to the input point.
        """
        close_pt = self.closest_point(point)
        return point.distance_to_point(close_pt)

    def intersect_line_ray(self, line_ray):
        """Get the intersection between this object and another Ray2 or LineSegment2D.

        Args:
            line_ray: Another LineSegment2D or Ray2D or to intersect.

        Returns:
            Point2D of intersection if it exists. None if no intersection exists.
        """
        return intersect_line2d(self, line_ray)

    def is_parallel(self, line_ray, angle_tolerance):
        """Test whether this object is parallel to another LineSegment2D or Ray2D.

        Args:
            line_ray: Another LineSegment2D or Ray2D for which parallelization
                with this objects will be tested.
            angle_tolerance: The max angle in radians that the direction between
                this object and another can vary for them to be considered
                parallel.
        """
        if self.v.angle(line_ray.v) <= angle_tolerance:
            return True
        elif self.v.angle(line_ray.v.reverse()) <= angle_tolerance:
            return True
        return False

    def is_colinear(self, line_ray, tolerance, angle_tolerance=None):
        """Test whether this object is colinear to another LineSegment2D or Ray2D.

        Args:
            line_ray: Another LineSegment2D or Ray2D for which co-linearity
                with this object will be tested.
            tolerance: The maximum distance between the line_ray and the infinite
                extension of this object for them to be considered colinear.
            angle_tolerance: The max angle in radians that the direction between
                this object and another can vary for them to be considered
                parallel. If None, the angle tolerance will not be used to
                evaluate co-linearity and the lines will only be considered
                colinear if the endpoints of one line are within the tolerance
                distance of the other line. (Default: None).
        """
        if angle_tolerance is not None and \
                not self.is_parallel(line_ray, angle_tolerance):
            return False
        _close_pt = closest_point2d_on_line2d_infinite(self.p, line_ray)
        if self.p.distance_to_point(_close_pt) >= tolerance:
            return False
        return True

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def to_dict(self):
        """Get LineSegment2D/Ray2D as a dictionary."""
        return {'p': self.p.to_array(),
                'v': self.v.to_array()}

    def __copy__(self):
        return self.__class__(self.p, self.v)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self.p), hash(self.v))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Base1DIn2D) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Base1Din2D representation."""
        return 'Base 1D Object (2D Space)'



# coding=utf-8
"""3D Mesh"""
from __future__ import division

from .._mesh import MeshBase
from ..util import coordinates_hash_3d
from ..geometry2d.mesh import Mesh2D
from .pointvector import Point3D, Vector3D
from .line import LineSegment3D
from .polyline import Polyline3D
from .plane import Plane

try:
    from itertools import izip as zip  # python 2
except ImportError:
    xrange = range  # python 3


class Mesh3D(MeshBase):
    """3D Mesh object.

    Args:
        vertices: A list or tuple of Point3D objects for vertices.
        faces: A list of tuples with each tuple having either 3 or 4 integers.
            These integers correspond to indices within the list of vertices.
        colors: An optional list of colors that correspond to either the faces
            of the mesh or the vertices of the mesh. Default is None.

    Properties:
        * vertices
        * faces
        * colors
        * is_color_by_face
        * min
        * max
        * center
        * area
        * face_areas
        * face_centroids
        * face_area_centroids
        * face_vertices
        * face_normals
        * vertex_normals
        * vertex_connected_faces
        * face_edges
        * edges
        * naked_edges
        * internal_edges
        * non_manifold_edges
    """
    __slots__ = ('_min', '_max', '_center', '_face_normals', '_vertex_normals')

    def __init__(self, vertices, faces, colors=None):
        """Initialize Mesh3D."""
        self._vertices = self._check_vertices_input(vertices)
        self._faces = self._check_faces_input(faces)

        self._is_color_by_face = False  # default if colors is None
        self.colors = colors
        self._min = None
        self._max = None
        self._center = None
        self._area = None
        self._face_areas = None
        self._face_centroids = None
        self._face_area_centroids = None
        self._face_normals = None
        self._vertex_normals = None
        self._vertex_connected_faces = None
        self._edge_indices = None
        self._edge_types = None
        self._edges = None
        self._naked_edges = None
        self._internal_edges = None
        self._non_manifold_edges = None

    @classmethod
    def from_dict(cls, data):
        """Create a Mesh3D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
                "type": "Mesh3D",
                "vertices": [(0, 0, 0), (10, 0, 0), (0, 10, 0)],
                "faces": [(0, 1, 2)],
                "colors": [{"r": 255, "g": 0, "b": 0}]
            }
        """
        colors = None
        if 'colors' in data and data['colors'] is not None and len(data['colors']) != 0:
            try:
                from ladybug.color import Color
            except ImportError:
                raise ImportError('Colors are specified in input Mesh2D dictionary '
                                  'but failed to import ladybug.color')
            colors = tuple(Color.from_dict(col) for col in data['colors'])
        fcs = tuple(tuple(f) for f in data['faces'])  # cast to immutable type
        return cls(tuple(Point3D.from_array(pt) for pt in data['vertices']), fcs, colors)

    @classmethod
    def from_face_vertices(cls, faces, purge=True):
        """Create a mesh from a list of faces with each face defined by Point3Ds.

        Args:
            faces: A list of faces with each face defined as a list of 3 or 4 Point3D.
            purge: A boolean to indicate if duplicate vertices should be shared between
                faces. Default is True to purge duplicate vertices, which can be slow
                for large lists of faces but results in a higher-quality mesh with
                a smaller size in memory. Note that vertices are only considered
                duplicate if the coordinate values are equal to one another
                within floating point tolerance. To remove duplicate vertices
                within a specified tolerance other than floating point, the
                from_purged_face_vertices method should be used instead.
        """
        vertices, face_collector = cls._interpret_input_from_face_vertices(faces, purge)
        return cls(tuple(vertices), tuple(face_collector))

    @classmethod
    def from_purged_face_vertices(cls, faces, tolerance):
        """Create a mesh from a list of faces with each face defined by Point3Ds.

        This method is slower than 'from_face_vertices' but will result in a mesh
        with fewer vertices and a smaller size in memory. This method is similar to
        using the 'purge' option in 'from_face_vertices' but will result in more shared
        vertices since it uses a tolerance to check equivalent vertices rather than
        comparing within floating point tolerance.

        Args:
            faces: A list of faces with each face defined as a list of 3 or 4 Point3D.
            tolerance: A number for the minimum difference between coordinate
                values at which point vertices are considered equal to one another.
        """
        vertices, faces = cls._interpret_input_from_face_vertices_with_tolerance(
            faces, tolerance)
        return cls(tuple(vertices), tuple(faces))

    @classmethod
    def from_mesh2d(cls, mesh_2d, plane=None):
        """Create a Mesh3D from a Mesh2D and a Plane in which the mesh exists.

        Args:
            mesh_2d: A Mesh2D object.
            plane: A Plane object to represent the plane in which the Mesh2D exists
                within 3D space. If None, the WorldXY plane will be used.
        """
        assert isinstance(mesh_2d, Mesh2D), 'Expected Mesh2D for from_mesh_2d. ' \
            'Got {}.'.format(type(mesh_2d))
        if plane is None:
            return cls(tuple(Point3D(pt.x, pt.y, 0) for pt in mesh_2d.vertices),
                       mesh_2d.faces, mesh_2d.colors)
        else:
            assert isinstance(plane, Plane), 'Expected Plane. Got {}'.format(type(plane))
            _verts3d = tuple(plane.xy_to_xyz(_v) for _v in mesh_2d.vertices)
            return cls(_verts3d, mesh_2d.faces, mesh_2d.colors)

    @classmethod
    def from_stl(cls, file_path):
        """Create a Mesh3D from an STL file.

        Args:
            file_path: Path to an STL file as a text string. The STL file can be
                in either ASCII or binary format.
        """
        from ladybug_geometry.interop.stl import STL  # avoid circular import
        face_vertices = STL.from_file(file_path).face_vertices
        return cls.from_face_vertices(face_vertices)

    @classmethod
    def from_obj(cls, file_path):
        """Create a Mesh3D from an OBJ file.

        Args:
            file_path: Path to an OBJ file as a text string.
        """
        from ladybug_geometry.interop.obj import OBJ  # avoid circular import
        transl_obj = OBJ.from_file(file_path)
        return cls(transl_obj.vertices, transl_obj.faces, transl_obj.vertex_colors)

    @property
    def min(self):
        """A Point3D for the minimum bounding box vertex around this mesh."""
        if self._min is None:
            self._calculate_min_max()
        return self._min

    @property
    def max(self):
        """A Point3D for the maximum bounding box vertex around this mesh."""
        if self._max is None:
            self._calculate_min_max()
        return self._max

    @property
    def center(self):
        """A Point3D for the center of the bounding box around this mesh."""
        if self._center is None:
            min, max = self.min, self.max
            self._center = Point3D(
                (min.x + max.x) / 2, (min.y + max.y) / 2, (min.z + max.z) / 2)
        return self._center

    @property
    def face_areas(self):
        """A tuple of face areas that parallels the faces property."""
        if self._face_normals is None:
            self._calculate_face_areas_and_normals()
        elif isinstance(self._face_areas, (float, int)):  # same area for each face
            self._face_areas = tuple(self._face_areas for face in self.faces)
        return self._face_areas

    @property
    def face_normals(self):
        """Tuple of Vector3D objects for all face normals."""
        if self._face_normals is None:
            self._calculate_face_areas_and_normals()
        elif isinstance(self._face_normals, Vector3D):  # same normal for each face
            self._face_normals = tuple(self._face_normals for face in self.faces)
        return self._face_normals

    @property
    def vertex_normals(self):
        """Tuple of Vector3D objects for all vertex normals."""
        if not self._vertex_normals:
            self._calculate_vertex_normals()
        elif isinstance(self._vertex_normals, Vector3D):  # same normal for each vertex
            self._vertex_normals = tuple(self._vertex_normals for face in self.vertices)
        return self._vertex_normals

    @property
    def face_edges(self):
        """List of polylines with one Polyline3D for each face.

        This is faster to compute compared to the edges and results in effectively
        the same type of wireframe visualization.
        """
        _all_verts = self._vertices
        f_edges = []
        for face in self._faces:
            verts = tuple(_all_verts[v] for v in face) + (_all_verts[face[0]],)
            f_edges.append(Polyline3D(verts))
        return f_edges

    @property
    def edges(self):
        """"Tuple of all edges in this Mesh3D as LineSegment3D objects.

        Note that this method will return only the unique edges in the mesh without
        any duplicates. This is sometimes desirable but can take a lot of time
        to compute for large meshes. For a faster property, use face_edges."""
        if self._edges is None:
            if self._edge_indices is None:
                self._compute_edge_info()
            self._edges = tuple(LineSegment3D.from_end_points(
                self.vertices[seg[0]], self.vertices[seg[1]])
                for seg in self._edge_indices)
        return self._edges

    @property
    def naked_edges(self):
        """"Tuple of all naked edges in this Mesh3D as LineSegment3D objects.

        Naked edges belong to only one face in the mesh (they are not
        shared between faces).
        """
        if self._naked_edges is None:
            self._naked_edges = self._get_edge_type(0)
        return self._naked_edges

    @property
    def internal_edges(self):
        """"Tuple of all internal edges in this Mesh3D as LineSegment3D objects.

        Internal edges are shared between two faces in the mesh.
        """
        if self._internal_edges is None:
            self._internal_edges = self._get_edge_type(1)
        return self._internal_edges

    @property
    def non_manifold_edges(self):
        """"Tuple of all non-manifold edges in this mesh as LineSegment3D objects.

        Non-manifold edges are shared between three or more faces.
        """
        if self._non_manifold_edges is None:
            if self._edges is None:
                self.edges
            nm_edges = []
            for i, type in enumerate(self._edge_types):
                if type > 1:
                    nm_edges.append(self._edges[i])
            self._non_manifold_edges = tuple(nm_edges)
        return self._non_manifold_edges

    def remove_vertices(self, pattern):
        """Get a version of this mesh where vertices are removed according to a pattern.

        Args:
            pattern: A list of boolean values denoting whether a vertex should
                remain in the mesh (True) or be removed from the mesh (False).
                The length of this list must match the number of this mesh's vertices.

        Returns:
            A tuple with two elements.

            -   new_mesh:
                A mesh where the vertices have been removed according
                to the input pattern.

            -   face_pattern:
                A list of boolean values that corresponds to the
                original mesh faces noting whether the face is in the new mesh
                (True) or has been removed from the new mesh (False).
        """
        _new_verts, _new_faces, _new_colors, _new_f_cent, _new_f_area, face_pattern = \
            self._remove_vertices(pattern)

        new_mesh = Mesh3D(_new_verts, _new_faces, _new_colors)
        new_mesh._face_centroids = _new_f_cent
        new_mesh._face_areas = _new_f_area
        return new_mesh, face_pattern

    def remove_faces(self, pattern):
        """Get a version of this mesh where faces are removed according to a pattern.

        Args:
            pattern: A list of boolean values denoting whether a face should
                remain in the mesh (True) or be removed from the mesh (False).
                The length of this list must match the number of this mesh's faces.

        Returns:
            A tuple with two elements.

            -   new_mesh:
                A mesh where the faces have been removed according
                to the input pattern.

            -   vertex_pattern:
                A list of boolean values that corresponds to the
                original mesh vertices noting whether the vertex is in the new mesh
                (True) or has been removed from the new mesh (False).
        """
        vertex_pattern = self._vertex_pattern_from_remove_faces(pattern)
        _new_verts, _new_faces, _new_colors, _new_f_cent, _new_f_area, face_pattern = \
            self._remove_vertices(vertex_pattern, pattern)

        new_mesh = Mesh3D(_new_verts, _new_faces, _new_colors)
        new_mesh._face_centroids = _new_f_cent
        new_mesh._face_areas = _new_f_area
        return new_mesh, vertex_pattern

    def remove_faces_only(self, pattern):
        """Get a version of this mesh where faces are removed and vertices are unaltered.

        This is faster than the Mesh3D.remove_faces method but will likely result
        a lower-quality mesh where several vertices exist in the mesh that are not
        referenced by any face. This may be preferred if pure speed of removing
        faces is a priority over smallest size of the mesh in memory.

        Args:
            pattern: A list of boolean values denoting whether a face should
                remain in the mesh (True) or be removed from the mesh (False).
                The length of this list must match the number of this mesh's faces.

        Returns:
            new_mesh -- A mesh where the faces have been removed according
            to the input pattern.
        """
        _new_faces, _new_colors, _new_f_cent, _new_f_area = \
            self._remove_faces_only(pattern)

        new_mesh = Mesh3D(self.vertices, _new_faces, _new_colors)
        new_mesh._face_centroids = _new_f_cent
        new_mesh._face_areas = _new_f_area
        return new_mesh

    def snap_to_points(self, points, distance):
        """Get a version of this mesh that is snapped to a set of points.

        Args:
            points: A list of Point3D to which the vertices of this mesh will be snapped.
            distance: The distance at which a mesh vertex close to a point will
                be snapped to it.
        """
        _new_verts = []
        for ev in self.vertices:
            dists = [ev.distance_to_point(pt) for pt in points]
            sort_pt = sorted(zip(dists, points), key=lambda pair: pair[0])
            if sort_pt[0][0] <= distance:
                _new_verts.append(sort_pt[0][1])
            else:
                _new_verts.append(ev)
        return Mesh3D(_new_verts, self.faces, self.colors)

    def apply_point_map(self, point_map, tolerance):
        """Apply a mapping of points from this mesh's vertices to a new mesh.

        Args:
            point_map: A dictionary with strings of coordinates as keys and Point3Ds
                as values. The keys refer to points within this Mesh3D to be
                moved to the new positions provided by the corresponding value.
                Correct keys can be generated by using the coordinates_hash_3d
                function on the vertices of this Mesh, found under the util module.
            tolerance: The minimum difference between point values that is
                considered meaningful. This is used as input to the
                coordinates_hash_3d function to generate correct hashes.
        """
        _new_verts = []
        for ev in self.vertices:
            try:
                _new_verts.append(point_map[coordinates_hash_3d(ev, tolerance)])
            except KeyError:  # a point that is not in the map
                _new_verts.append(ev)
        return Mesh3D(_new_verts, self.faces, self.colors)

    def rotate(self, axis, angle, origin):
        """Rotate a mesh by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the point will be rotated.
        """
        _verts = tuple(pt.rotate(axis, angle, origin) for pt in self.vertices)
        return self._mesh_transform(_verts)

    def rotate_xy(self, angle, origin):
        """Get a mesh rotated counterclockwise in the XY plane by a certain angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the point will be rotated.
        """
        _verts = tuple(pt.rotate_xy(angle, origin) for pt in self.vertices)
        return self._mesh_transform(_verts)

    def scale(self, factor, origin=None):
        """Scale a mesh by a factor from an origin point.

        Args:
            factor: A number representing how much the mesh should be scaled.
            origin: A Point representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        if origin is None:
            _verts = tuple(
                Point3D(pt.x * factor, pt.y * factor, pt.z * factor)
                for pt in self.vertices)
        else:
            _verts = tuple(pt.scale(factor, origin) for pt in self.vertices)
        return self._mesh_scale(_verts, factor)

    def offset_mesh(self, distance):
        """Get a Mesh3D that has been offset from this one by a certain difference.

        Effectively, this method moves each mesh vertex along the vertex normal
        by the offset distance.

        Args:
            distance: A number for the distance to offset the mesh.
        """
        new_verts = tuple(pt.move(norm * distance) for pt, norm in
                          zip(self.vertices, self.vertex_normals))
        return Mesh3D(new_verts, self.faces, self._colors)

    def height_field_mesh(self, values, domain):
        """Get a Mesh3D that has faces or vertices offset according to a list of values.

        Args:
            values: A list of values that has a length matching the number of faces
                or vertices in this mesh.
            domain: A tuple or list of two numbers for the upper and lower distances
                that the mesh vertices should be offset. (ie. (0, 3))
        """
        assert isinstance(domain, (tuple, list)), 'Expected tuple for domain. '\
            'Got {}.'.format(type(domain))
        assert len(domain) == 2, 'Expected domain to be in the format (min, max). ' \
            'Got {}.'.format(domain)

        if len(values) == len(self.faces):
            remap_vals = Mesh3D._remap_values(values, domain[0], domain[-1])
            vert_remap_vals = []
            for vf in self.vertex_connected_faces:
                v = 0
                for j in vf:
                    v += remap_vals[j]
                try:
                    v /= len(vf)  # average the vertex value over its connected faces
                except ZeroDivisionError:
                    pass  # lone vertex without any faces
                vert_remap_vals.append(v)
            new_verts = tuple(pt.move(norm * dist) for pt, norm, dist in
                              zip(self.vertices, self.vertex_normals, vert_remap_vals))
        elif len(values) == len(self.vertices):
            remap_vals = Mesh3D._remap_values(values, domain[0], domain[-1])
            new_verts = tuple(pt.move(norm * dist) for pt, norm, dist in
                              zip(self.vertices, self.vertex_normals, remap_vals))
        else:
            raise ValueError(
                'Input values for height_field_mesh ({}) does not match the number of'
                ' mesh faces ({}) nor the number of vertices ({}).'
                .format(len(values), len(self.faces), len(self.vertices)))
        return Mesh3D(new_verts, self.faces, self._colors)

    def to_dict(self):
        """Get Mesh3D as a dictionary."""
        base = {'type': 'Mesh3D',
                'vertices': [pt.to_array() for pt in self.vertices],
                'faces': self.faces}
        if self.colors is not None:
            base['colors'] = [col.to_dict() for col in self.colors]
        return base

    def to_stl(self, folder, name=None):
        """Write the Mesh3D to an ASCII STL file.

        Args:
            folder: A text string for the directory where the STL will be written.
            name: A text string for the name of the STL file.
        """
        from ladybug_geometry.interop.stl import STL  # avoid circular import
        stl_obj = STL(self.face_vertices, self.face_normals)
        return stl_obj.to_file(folder, name)

    def to_obj(self, folder, name, include_colors=True, include_normals=False,
               triangulate_quads=False, include_mtl=False):
        """Write the Mesh3D to an ASCII OBJ file.

        Args:
            folder: A text string for the directory where the OBJ will be written.
            name: A text string for the name of the OBJ file.
            include_colors: Boolean to note whether the Mesh3D colors should be
                included in the OBJ file. (Default: True).
            include_normals: Boolean to note whether the vertex normals should be
                included in the OBJ file. (Default: False).
            triangulate_quads: Boolean to note whether quad faces should be
                triangulated upon export to OBJ. This may be needed for certain
                software platforms that require the mesh to be composed entirely
                of triangles (eg. Radiance). (Default: False).
            include_mtl: Boolean to note whether an .mtl file should be automatically
                generated next to the .obj file in the output folder. All materials
                in the mtl file will be diffuse white, with the assumption that
                these will be customized later. (Default: False).
        """
        from ladybug_geometry.interop.obj import OBJ  # avoid circular import
        transl_obj = OBJ.from_mesh3d(self, include_colors, include_normals)
        return transl_obj.to_file(folder, name, triangulate_quads, include_mtl)

    @staticmethod
    def join_meshes(meshes):
        """Join an array of Mesh3Ds into a single Mesh3D.

        Args:
            meshes: An array of meshes to be joined into one.

        Returns:
            A single Mesh3D object derived from the input meshes.
        """
        # set up empty lists of objects to be filled
        verts = []
        faces = []
        colors = []

        # loop through all of the meshes and get new faces
        total_v_i = 0
        for mesh in meshes:
            verts.extend(mesh._vertices)
            for fc in mesh._faces:
                faces.append(tuple(v_i + total_v_i for v_i in fc))
            total_v_i += len(mesh._vertices)
            if mesh._colors:
                colors.extend(mesh._colors)

        # create the new mesh
        if len(colors) != 0:
            new_mesh = Mesh3D(verts, faces, colors)
        else:
            new_mesh = Mesh3D(verts, faces)
        return new_mesh

    def _calculate_min_max(self):
        """Calculate maximum and minimum Point3D for this object."""
        min_pt = [self.vertices[0].x, self.vertices[0].y, self.vertices[0].z]
        max_pt = [self.vertices[0].x, self.vertices[0].y, self.vertices[0].z]

        for v in self.vertices[1:]:
            if v.x < min_pt[0]:
                min_pt[0] = v.x
            elif v.x > max_pt[0]:
                max_pt[0] = v.x
            if v.y < min_pt[1]:
                min_pt[1] = v.y
            elif v.y > max_pt[1]:
                max_pt[1] = v.y
            if v.z < min_pt[2]:
                min_pt[2] = v.z
            elif v.z > max_pt[2]:
                max_pt[2] = v.z

        self._min = Point3D(min_pt[0], min_pt[1], min_pt[2])
        self._max = Point3D(max_pt[0], max_pt[1], max_pt[2])

    def _calculate_face_areas_and_normals(self):
        """Calculate face areas and normals from vertices."""
        _f_norm = []
        _f_area = []
        for face in self.faces:
            pts = tuple(self._vertices[i] for i in face)
            if len(face) == 3:
                n, a = self._calculate_normal_and_area_for_triangle(pts)
            else:
                n, a = self._calculate_normal_and_area_for_quad(pts)
            _f_norm.append(n)
            _f_area.append(a)
        self._face_normals = tuple(_f_norm)
        self._face_areas = tuple(_f_area)

    def _calculate_vertex_normals(self):
        """Calculate vertex normals.

        This is accomplished by normalizing the average of the surface normals
        of the faces that contain that vertex.  This particular method weights
        this average by the area of each face, though this does not always need
        to be the case as noted here:
        https://en.wikipedia.org/wiki/Vertex_normal
        """
        # find shared faces for each vertices
        mapper = [[] for v in xrange(len(self.vertices))]
        for c, face in enumerate(self.faces):
            for i in face:
                mapper[i].append(c)
        # now calculate vertex normal based on face normals
        vn = []
        fn = self.face_normals
        fa = self.face_areas
        for fi in mapper:
            x, y, z = 0, 0, 0
            for n, a in zip(tuple(fn[i] for i in fi), tuple(fa[i] for i in fi)):
                x += n.x * a
                y += n.y * a
                z += n.z * a
            _v = Vector3D(x, y, z)
            vn.append(_v.normalize())
        self._vertex_normals = tuple(vn)

    def _get_edge_type(self, edge_type):
        """Get all of the edges of a certain type in this mesh."""
        if self._edges is None:
            self.edges
        sel_edges = []
        for i, type in enumerate(self._edge_types):
            if type == edge_type:
                sel_edges.append(self._edges[i])
        return tuple(sel_edges)

    def _tri_face_centroid(self, face):
        """Compute the centroid of a triangular face."""
        return Mesh3D._tri_centroid(tuple(self._vertices[i] for i in face))

    def _quad_face_centroid(self, face):
        """Compute the centroid of a quadrilateral face."""
        return Mesh3D._quad_centroid(tuple(self._vertices[i] for i in face))

    def _mesh_transform(self, verts):
        """Transform mesh in a way that transfers properties and avoids extra checks."""
        _new_mesh = Mesh3D(verts, self.faces)
        self._transfer_properties(_new_mesh)
        return _new_mesh

    def _mesh_transform_move(self, verts):
        """Move mesh in a way that transfers properties and avoids extra checks."""
        _new_mesh = Mesh3D(verts, self.faces)
        self._transfer_properties(_new_mesh)
        _new_mesh._face_normals = self._face_normals
        _new_mesh._vertex_normals = self._vertex_normals
        return _new_mesh

    def _mesh_scale(self, verts, factor):
        """Scale mesh in a way that transfers properties and avoids extra checks."""
        _new_mesh = Mesh3D(verts, self.faces)
        self._transfer_properties_scale(_new_mesh, factor)
        _new_mesh._face_normals = self._face_normals
        _new_mesh._vertex_normals = self._vertex_normals
        return _new_mesh

    def _check_vertices_input(self, vertices):
        """Check the input vertices."""
        if not isinstance(vertices, tuple):
            vertices = tuple(vertices)
        for vert in vertices:
            assert isinstance(vert, Point3D), \
                'Expected Point3D for {} vertex. Got {}.'.format(
                    self.__class__.__name__, type(vert))
        return vertices

    @staticmethod
    def _calculate_normal_and_area_for_triangle(pts):
        """Calculate normal and area for three points.

        Returns:
            n = Normalized normal vector for the triangle.
            a = Area of the triangle.
        """
        v1 = pts[1] - pts[0]
        v2 = pts[2] - pts[0]
        n = v1.cross(v2)
        a = n.magnitude / 2
        return n.normalize(), a

    @staticmethod
    def _calculate_normal_and_area_for_quad(pts):
        """Calculate normal and area for four points.

        This method uses an area-weighted average of the two triangle normals
        that compose the quad face.

        Returns:
            n = Normalized normal vector for the quad.
            a = Area of the quad.
        """
        # TODO: Get this method to work for concave quads.
        # This method is only reliable when quads are convex since we assume
        # either diagonal of the quad splits it into two triangles.
        # It seems Rhino never produces concave quads when it automatically meshes
        # but we will likely want to add support for this if meshes have other origins
        v1 = pts[1] - pts[0]
        v2 = pts[2] - pts[0]
        n1 = v1.cross(v2)

        v3 = pts[3] - pts[2]
        v4 = pts[1] - pts[2]
        n2 = v3.cross(v4)

        a = (n1.magnitude + n2.magnitude) / 2
        n = Vector3D((n1.x + n2.x) / 2, (n1.y + n2.y) / 2, (n1.z + n2.z) / 2)
        return n.normalize(), a

    @staticmethod
    def _face_center(verts):
        """Get the center of a list of Point3D vertices."""
        _cent_x = sum([v.x for v in verts])
        _cent_y = sum([v.y for v in verts])
        _cent_z = sum([v.z for v in verts])
        v_count = len(verts)
        return Point3D(_cent_x / v_count, _cent_y / v_count, _cent_z / v_count)

    @staticmethod
    def _tri_centroid(verts):
        """Get the centroid of a list of 3 Point3D vertices."""
        _cent_x = sum([v.x for v in verts])
        _cent_y = sum([v.y for v in verts])
        _cent_z = sum([v.z for v in verts])
        return Point3D(_cent_x / 3, _cent_y / 3, _cent_z / 3)

    @staticmethod
    def _quad_centroid(verts):
        """Get the centroid of a list of 4 Point3D vertices."""
        # TODO: Get this method to recognize concave quads.
        # This method is only reliable when quads are convex since we assume
        # either diagonal of the quad splits it into two triangles.
        # It seems Rhino never produces concave quads when it automatically meshes
        _tri_verts = ((verts[0], verts[1], verts[2]), (verts[2], verts[3], verts[0]))
        _tri_c = [Mesh3D._tri_centroid(tri) for tri in _tri_verts]
        _tri_a = [Mesh3D._get_tri_area(tri) for tri in _tri_verts]
        _tot_a = sum(_tri_a)
        try:
            _cent_x = (_tri_c[0].x * _tri_a[0] + _tri_c[1].x * _tri_a[1]) / _tot_a
            _cent_y = (_tri_c[0].y * _tri_a[0] + _tri_c[1].y * _tri_a[1]) / _tot_a
            _cent_z = (_tri_c[0].z * _tri_a[0] + _tri_c[1].z * _tri_a[1]) / _tot_a
        except ZeroDivisionError:
            _cent_x = sum([v.x for v in verts]) / 4
            _cent_y = sum([v.y for v in verts]) / 4
            _cent_z = sum([v.z for v in verts]) / 4
        return Point3D(_cent_x, _cent_y, _cent_z)

    @staticmethod
    def _get_tri_area(pts):
        """Get the area of a triangle from three Point3D objects."""
        v1 = pts[1] - pts[0]
        v2 = pts[2] - pts[0]
        n1 = v1.cross(v2)
        return n1.magnitude / 2

    @staticmethod
    def _remap_values(values, tmin, tmax):
        """Remap a set of values to offset distances within a domain."""
        omin = min(values)
        omax = max(values)
        odiff = omax - omin
        tdiff = tmax - tmin
        if odiff == 0:
            return [tmin] * len(values)
        else:
            return [(v - omin) * tdiff / odiff + tmin for v in values]

    def __copy__(self):
        _new_mesh = Mesh3D(self.vertices, self.faces)
        self._transfer_properties(_new_mesh)
        _new_mesh._face_centroids = self._face_centroids
        _new_mesh._face_normals = self._face_normals
        _new_mesh._vertex_normals = self._vertex_normals
        return _new_mesh

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return tuple(hash(pt) for pt in self._vertices) + \
            tuple(hash(face) for face in self._faces)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Mesh3D) and self.__key() == other.__key()

    def __repr__(self):
        return 'Mesh3D ({} faces) ({} vertices)'.format(
            len(self.faces), len(self))



"""Geometry objects in 2D space."""

from .pointvector import Vector2D, Point2D
from .ray import Ray2D
from .line import LineSegment2D
from .arc import Arc2D
from .polyline import Polyline2D
from .polygon import Polygon2D
from .mesh import Mesh2D



"""Utility functions used by several different geometry methods."""
from __future__ import division
import math


def coordinates_hash(point, tolerance):
    """Convert XY coordinates of a Point2D into a string useful for hashing.

    Points that are co-located within the tolerance will receive the same string value
    from this function, which helps convert line segments that contain duplicated
    vertex references them into a singular network object where co-located vertices
    are referenced only once.

    Args:
        point: A Point2D object.
        tolerance: floating point precision tolerance.

    Returns:
        A string of rounded coordinates.
    """
    # get the relative tolerance using a log function
    try:
        rtol = int(math.log10(tolerance)) * -1
    except ValueError:
        rtol = 0  # the tol is equal to 1 (out of range for log)
    # account for the fact that the tolerance may not be base 10
    base = int(tolerance * 10 ** (rtol + 1))
    if base == 10 or base == 0:  # tolerance is base 10 (eg. 0.001)
        base = 1
    else:  # tolerance is not base 10 (eg. 0.003)
        rtol += 1
    # avoid cases of signed zeros messing with the hash
    z_tol = tolerance / 2
    x_val = 0.0 if abs(point.x) < z_tol else point.x
    y_val = 0.0 if abs(point.y) < z_tol else point.y
    # convert the coordinate values to a hash
    return str((
        base * round(x_val / base, rtol),
        base * round(y_val / base, rtol)
    ))


def coordinates_hash_3d(point, tolerance):
    """Convert XY coordinates of a Point3D into a string useful for hashing.

    Points that are co-located within the tolerance will receive the same string value
    from this function, which helps convert line segments that contain duplicated
    vertex references them into a singular network object where co-located vertices
    are referenced only once.

    Args:
        point: A Point3D object.
        tolerance: floating point precision tolerance.

    Returns:
        A string of rounded coordinates.
    """
    # get the relative tolerance using a log function
    try:
        rtol = int(math.log10(tolerance)) * -1
    except ValueError:
        rtol = 0  # the tol is equal to 1 (out of range for log)
    # account for the fact that the tolerance may not be base 10
    base = int(tolerance * 10 ** (rtol + 1))
    if base == 10 or base == 0:  # tolerance is base 10 (eg. 0.001)
        base = 1
    else:  # tolerance is not base 10 (eg. 0.003)
        rtol += 1
    # avoid cases of signed zeros messing with the hash
    z_tol = tolerance / 2
    x_val = 0.0 if abs(point.x) < z_tol else point.x
    y_val = 0.0 if abs(point.y) < z_tol else point.y
    z_val = 0.0 if abs(point.z) < z_tol else point.z
    # convert the coordinate values to a hash
    return str((
        base * round(x_val / base, rtol),
        base * round(y_val / base, rtol),
        base * round(z_val / base, rtol)
    ))
